# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: fonts_public.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='fonts_public.proto',
  package='google.fonts_public',
  syntax='proto2',
  serialized_options=b'\n\026com.google.fonts.protoB\013FontsPublic',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12\x66onts_public.proto\x12\x13google.fonts_public\"\x89\x06\n\x0b\x46\x61milyProto\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x10\n\x08\x64\x65signer\x18\x02 \x02(\t\x12\x0f\n\x07license\x18\x03 \x02(\t\x12\x10\n\x08\x63\x61tegory\x18\x04 \x03(\t\x12\x12\n\ndate_added\x18\x05 \x02(\t\x12-\n\x05\x66onts\x18\x06 \x03(\x0b\x32\x1e.google.fonts_public.FontProto\x12\x0f\n\x07\x61liases\x18\x07 \x03(\t\x12\x0f\n\x07subsets\x18\x08 \x03(\t\x12\x19\n\x11ttf_autohint_args\x18\t \x01(\t\x12\x33\n\x04\x61xes\x18\n \x03(\x0b\x32%.google.fonts_public.AxisSegmentProto\x12\x62\n\x1aregistry_default_overrides\x18\x0b \x03(\x0b\x32>.google.fonts_public.FamilyProto.RegistryDefaultOverridesEntry\x12\x30\n\x06source\x18\x0c \x01(\x0b\x32 .google.fonts_public.SourceProto\x12\x0f\n\x07is_noto\x18\r \x01(\x08\x12\x11\n\tlanguages\x18\x0e \x03(\t\x12;\n\tfallbacks\x18\x0f \x03(\x0b\x32(.google.fonts_public.FamilyFallbackProto\x12;\n\rsample_glyphs\x18\x10 \x03(\x0b\x32$.google.fonts_public.GlyphGroupProto\x12\x39\n\x0bsample_text\x18\x11 \x01(\x0b\x32$.google.fonts_public.SampleTextProto\x12\x14\n\x0c\x64isplay_name\x18\x12 \x01(\t\x12<\n\x0bsource_type\x18\x13 \x01(\x0e\x32\x1d.google.fonts_public.FontType:\x08TYPE_TTF\x1a?\n\x1dRegistryDefaultOverridesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\"\x8a\x01\n\tFontProto\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05style\x18\x02 \x02(\t\x12\x0e\n\x06weight\x18\x03 \x02(\x05\x12\x10\n\x08\x66ilename\x18\x04 \x02(\t\x12\x18\n\x10post_script_name\x18\x05 \x02(\t\x12\x11\n\tfull_name\x18\x06 \x02(\t\x12\x11\n\tcopyright\x18\x07 \x01(\t\"Z\n\x10\x41xisSegmentProto\x12\x0b\n\x03tag\x18\x01 \x01(\t\x12\x11\n\tmin_value\x18\x02 \x01(\x02\x12\x11\n\tmax_value\x18\x04 \x01(\x02J\x04\x08\x03\x10\x04R\rdefault_value\"5\n\x0bSourceProto\x12\x16\n\x0erepository_url\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\t\"H\n\x0bTargetProto\x12\x39\n\x0btarget_type\x18\x01 \x01(\x0e\x32$.google.fonts_public.TargetTypeProto\"\xcc\x01\n\x13\x46\x61milyFallbackProto\x12:\n\x0b\x61xis_target\x18\x01 \x03(\x0b\x32%.google.fonts_public.AxisSegmentProto\x12\x30\n\x06target\x18\x02 \x03(\x0b\x32 .google.fonts_public.TargetProto\x12\x17\n\x0fsize_adjust_pct\x18\x03 \x01(\x02\x12\x1b\n\x13\x61scent_override_pct\x18\x05 \x01(\x02\x12\x11\n\tlocal_src\x18\x04 \x03(\t\"Q\n\x0bRegionProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\npopulation\x18\x03 \x01(\x05\x12\x14\n\x0cregion_group\x18\x04 \x03(\t\"\'\n\x0bScriptProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"\xa8\x02\n\rLanguageProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08language\x18\x02 \x01(\t\x12\x0e\n\x06script\x18\x03 \x01(\t\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\x16\n\x0epreferred_name\x18\x05 \x01(\t\x12\x0f\n\x07\x61utonym\x18\x06 \x01(\t\x12\x12\n\npopulation\x18\x07 \x01(\x05\x12\x0e\n\x06region\x18\x08 \x03(\t\x12?\n\x0e\x65xemplar_chars\x18\t \x01(\x0b\x32\'.google.fonts_public.ExemplarCharsProto\x12\x39\n\x0bsample_text\x18\n \x01(\x0b\x32$.google.fonts_public.SampleTextProto\x12\x12\n\nhistorical\x18\x0b \x01(\x08\"z\n\x12\x45xemplarCharsProto\x12\x0c\n\x04\x62\x61se\x18\x01 \x01(\t\x12\x11\n\tauxiliary\x18\x02 \x01(\t\x12\r\n\x05marks\x18\x03 \x01(\t\x12\x10\n\x08numerals\x18\x04 \x01(\t\x12\x13\n\x0bpunctuation\x18\x05 \x01(\t\x12\r\n\x05index\x18\x06 \x01(\t\"\x84\x02\n\x0fSampleTextProto\x12\x15\n\rmasthead_full\x18\x01 \x01(\t\x12\x18\n\x10masthead_partial\x18\x02 \x01(\t\x12\x0e\n\x06styles\x18\x03 \x01(\t\x12\x0e\n\x06tester\x18\x04 \x01(\t\x12\x11\n\tposter_sm\x18\x05 \x01(\t\x12\x11\n\tposter_md\x18\x06 \x01(\t\x12\x11\n\tposter_lg\x18\x07 \x01(\t\x12\x13\n\x0bspecimen_48\x18\x08 \x01(\t\x12\x13\n\x0bspecimen_36\x18\t \x01(\t\x12\x13\n\x0bspecimen_32\x18\n \x01(\t\x12\x13\n\x0bspecimen_21\x18\x0b \x01(\t\x12\x13\n\x0bspecimen_16\x18\x0c \x01(\t\"/\n\x0fGlyphGroupProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06glyphs\x18\x02 \x01(\t*\x92\x01\n\x0fTargetTypeProto\x12\x16\n\x12TARGET_UNSPECIFIED\x10\x00\x12\x15\n\x11TARGET_OS_WINDOWS\x10\x01\x12\x11\n\rTARGET_OS_MAC\x10\x02\x12\x13\n\x0fTARGET_OS_LINUX\x10\x03\x12\x15\n\x11TARGET_OS_ANDROID\x10\x04\x12\x11\n\rTARGET_OS_IOS\x10\x05*o\n\x08\x46ontType\x12\x0c\n\x08TYPE_TTF\x10\x00\x12\x0c\n\x08TYPE_EOT\x10\x01\x12\x0c\n\x08TYPE_SVG\x10\x02\x12\r\n\tTYPE_WOFF\x10\x03\x12\x0e\n\nTYPE_WOFF2\x10\x05\x12\x0c\n\x08TYPE_SWF\x10\x04\x12\x0c\n\x08TYPE_OTF\x10\x06\x42%\n\x16\x63om.google.fonts.protoB\x0b\x46ontsPublic'
)

_TARGETTYPEPROTO = _descriptor.EnumDescriptor(
  name='TargetTypeProto',
  full_name='google.fonts_public.TargetTypeProto',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TARGET_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TARGET_OS_WINDOWS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TARGET_OS_MAC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TARGET_OS_LINUX', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TARGET_OS_ANDROID', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TARGET_OS_IOS', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2252,
  serialized_end=2398,
)
_sym_db.RegisterEnumDescriptor(_TARGETTYPEPROTO)

TargetTypeProto = enum_type_wrapper.EnumTypeWrapper(_TARGETTYPEPROTO)
_FONTTYPE = _descriptor.EnumDescriptor(
  name='FontType',
  full_name='google.fonts_public.FontType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_TTF', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_EOT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_SVG', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_WOFF', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_WOFF2', index=4, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_SWF', index=5, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_OTF', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2400,
  serialized_end=2511,
)
_sym_db.RegisterEnumDescriptor(_FONTTYPE)

FontType = enum_type_wrapper.EnumTypeWrapper(_FONTTYPE)
TARGET_UNSPECIFIED = 0
TARGET_OS_WINDOWS = 1
TARGET_OS_MAC = 2
TARGET_OS_LINUX = 3
TARGET_OS_ANDROID = 4
TARGET_OS_IOS = 5
TYPE_TTF = 0
TYPE_EOT = 1
TYPE_SVG = 2
TYPE_WOFF = 3
TYPE_WOFF2 = 5
TYPE_SWF = 4
TYPE_OTF = 6



_FAMILYPROTO_REGISTRYDEFAULTOVERRIDESENTRY = _descriptor.Descriptor(
  name='RegistryDefaultOverridesEntry',
  full_name='google.fonts_public.FamilyProto.RegistryDefaultOverridesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='google.fonts_public.FamilyProto.RegistryDefaultOverridesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='google.fonts_public.FamilyProto.RegistryDefaultOverridesEntry.value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=758,
  serialized_end=821,
)

_FAMILYPROTO = _descriptor.Descriptor(
  name='FamilyProto',
  full_name='google.fonts_public.FamilyProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts_public.FamilyProto.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='designer', full_name='google.fonts_public.FamilyProto.designer', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='license', full_name='google.fonts_public.FamilyProto.license', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='category', full_name='google.fonts_public.FamilyProto.category', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date_added', full_name='google.fonts_public.FamilyProto.date_added', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fonts', full_name='google.fonts_public.FamilyProto.fonts', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='aliases', full_name='google.fonts_public.FamilyProto.aliases', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subsets', full_name='google.fonts_public.FamilyProto.subsets', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ttf_autohint_args', full_name='google.fonts_public.FamilyProto.ttf_autohint_args', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='axes', full_name='google.fonts_public.FamilyProto.axes', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registry_default_overrides', full_name='google.fonts_public.FamilyProto.registry_default_overrides', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='google.fonts_public.FamilyProto.source', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_noto', full_name='google.fonts_public.FamilyProto.is_noto', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='languages', full_name='google.fonts_public.FamilyProto.languages', index=13,
      number=14, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fallbacks', full_name='google.fonts_public.FamilyProto.fallbacks', index=14,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_glyphs', full_name='google.fonts_public.FamilyProto.sample_glyphs', index=15,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_text', full_name='google.fonts_public.FamilyProto.sample_text', index=16,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='google.fonts_public.FamilyProto.display_name', index=17,
      number=18, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_type', full_name='google.fonts_public.FamilyProto.source_type', index=18,
      number=19, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_FAMILYPROTO_REGISTRYDEFAULTOVERRIDESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=821,
)


_FONTPROTO = _descriptor.Descriptor(
  name='FontProto',
  full_name='google.fonts_public.FontProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts_public.FontProto.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='style', full_name='google.fonts_public.FontProto.style', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='weight', full_name='google.fonts_public.FontProto.weight', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filename', full_name='google.fonts_public.FontProto.filename', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='post_script_name', full_name='google.fonts_public.FontProto.post_script_name', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='full_name', full_name='google.fonts_public.FontProto.full_name', index=5,
      number=6, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='copyright', full_name='google.fonts_public.FontProto.copyright', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=824,
  serialized_end=962,
)


_AXISSEGMENTPROTO = _descriptor.Descriptor(
  name='AxisSegmentProto',
  full_name='google.fonts_public.AxisSegmentProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tag', full_name='google.fonts_public.AxisSegmentProto.tag', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_value', full_name='google.fonts_public.AxisSegmentProto.min_value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_value', full_name='google.fonts_public.AxisSegmentProto.max_value', index=2,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=964,
  serialized_end=1054,
)


_SOURCEPROTO = _descriptor.Descriptor(
  name='SourceProto',
  full_name='google.fonts_public.SourceProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='repository_url', full_name='google.fonts_public.SourceProto.repository_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='commit', full_name='google.fonts_public.SourceProto.commit', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1056,
  serialized_end=1109,
)


_TARGETPROTO = _descriptor.Descriptor(
  name='TargetProto',
  full_name='google.fonts_public.TargetProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='target_type', full_name='google.fonts_public.TargetProto.target_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1111,
  serialized_end=1183,
)


_FAMILYFALLBACKPROTO = _descriptor.Descriptor(
  name='FamilyFallbackProto',
  full_name='google.fonts_public.FamilyFallbackProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis_target', full_name='google.fonts_public.FamilyFallbackProto.axis_target', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target', full_name='google.fonts_public.FamilyFallbackProto.target', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size_adjust_pct', full_name='google.fonts_public.FamilyFallbackProto.size_adjust_pct', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ascent_override_pct', full_name='google.fonts_public.FamilyFallbackProto.ascent_override_pct', index=3,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='local_src', full_name='google.fonts_public.FamilyFallbackProto.local_src', index=4,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1186,
  serialized_end=1390,
)


_REGIONPROTO = _descriptor.Descriptor(
  name='RegionProto',
  full_name='google.fonts_public.RegionProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='google.fonts_public.RegionProto.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts_public.RegionProto.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='population', full_name='google.fonts_public.RegionProto.population', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='region_group', full_name='google.fonts_public.RegionProto.region_group', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1392,
  serialized_end=1473,
)


_SCRIPTPROTO = _descriptor.Descriptor(
  name='ScriptProto',
  full_name='google.fonts_public.ScriptProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='google.fonts_public.ScriptProto.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts_public.ScriptProto.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1475,
  serialized_end=1514,
)


_LANGUAGEPROTO = _descriptor.Descriptor(
  name='LanguageProto',
  full_name='google.fonts_public.LanguageProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='google.fonts_public.LanguageProto.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='language', full_name='google.fonts_public.LanguageProto.language', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='script', full_name='google.fonts_public.LanguageProto.script', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts_public.LanguageProto.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='preferred_name', full_name='google.fonts_public.LanguageProto.preferred_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='autonym', full_name='google.fonts_public.LanguageProto.autonym', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='population', full_name='google.fonts_public.LanguageProto.population', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='region', full_name='google.fonts_public.LanguageProto.region', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exemplar_chars', full_name='google.fonts_public.LanguageProto.exemplar_chars', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_text', full_name='google.fonts_public.LanguageProto.sample_text', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='historical', full_name='google.fonts_public.LanguageProto.historical', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1517,
  serialized_end=1813,
)


_EXEMPLARCHARSPROTO = _descriptor.Descriptor(
  name='ExemplarCharsProto',
  full_name='google.fonts_public.ExemplarCharsProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='google.fonts_public.ExemplarCharsProto.base', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auxiliary', full_name='google.fonts_public.ExemplarCharsProto.auxiliary', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='marks', full_name='google.fonts_public.ExemplarCharsProto.marks', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='numerals', full_name='google.fonts_public.ExemplarCharsProto.numerals', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='punctuation', full_name='google.fonts_public.ExemplarCharsProto.punctuation', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index', full_name='google.fonts_public.ExemplarCharsProto.index', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1815,
  serialized_end=1937,
)


_SAMPLETEXTPROTO = _descriptor.Descriptor(
  name='SampleTextProto',
  full_name='google.fonts_public.SampleTextProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='masthead_full', full_name='google.fonts_public.SampleTextProto.masthead_full', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='masthead_partial', full_name='google.fonts_public.SampleTextProto.masthead_partial', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='styles', full_name='google.fonts_public.SampleTextProto.styles', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tester', full_name='google.fonts_public.SampleTextProto.tester', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='poster_sm', full_name='google.fonts_public.SampleTextProto.poster_sm', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='poster_md', full_name='google.fonts_public.SampleTextProto.poster_md', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='poster_lg', full_name='google.fonts_public.SampleTextProto.poster_lg', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='specimen_48', full_name='google.fonts_public.SampleTextProto.specimen_48', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='specimen_36', full_name='google.fonts_public.SampleTextProto.specimen_36', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='specimen_32', full_name='google.fonts_public.SampleTextProto.specimen_32', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='specimen_21', full_name='google.fonts_public.SampleTextProto.specimen_21', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='specimen_16', full_name='google.fonts_public.SampleTextProto.specimen_16', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1940,
  serialized_end=2200,
)


_GLYPHGROUPPROTO = _descriptor.Descriptor(
  name='GlyphGroupProto',
  full_name='google.fonts_public.GlyphGroupProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts_public.GlyphGroupProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='glyphs', full_name='google.fonts_public.GlyphGroupProto.glyphs', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2202,
  serialized_end=2249,
)

_FAMILYPROTO_REGISTRYDEFAULTOVERRIDESENTRY.containing_type = _FAMILYPROTO
_FAMILYPROTO.fields_by_name['fonts'].message_type = _FONTPROTO
_FAMILYPROTO.fields_by_name['axes'].message_type = _AXISSEGMENTPROTO
_FAMILYPROTO.fields_by_name['registry_default_overrides'].message_type = _FAMILYPROTO_REGISTRYDEFAULTOVERRIDESENTRY
_FAMILYPROTO.fields_by_name['source'].message_type = _SOURCEPROTO
_FAMILYPROTO.fields_by_name['fallbacks'].message_type = _FAMILYFALLBACKPROTO
_FAMILYPROTO.fields_by_name['sample_glyphs'].message_type = _GLYPHGROUPPROTO
_FAMILYPROTO.fields_by_name['sample_text'].message_type = _SAMPLETEXTPROTO
_FAMILYPROTO.fields_by_name['source_type'].enum_type = _FONTTYPE
_TARGETPROTO.fields_by_name['target_type'].enum_type = _TARGETTYPEPROTO
_FAMILYFALLBACKPROTO.fields_by_name['axis_target'].message_type = _AXISSEGMENTPROTO
_FAMILYFALLBACKPROTO.fields_by_name['target'].message_type = _TARGETPROTO
_LANGUAGEPROTO.fields_by_name['exemplar_chars'].message_type = _EXEMPLARCHARSPROTO
_LANGUAGEPROTO.fields_by_name['sample_text'].message_type = _SAMPLETEXTPROTO
DESCRIPTOR.message_types_by_name['FamilyProto'] = _FAMILYPROTO
DESCRIPTOR.message_types_by_name['FontProto'] = _FONTPROTO
DESCRIPTOR.message_types_by_name['AxisSegmentProto'] = _AXISSEGMENTPROTO
DESCRIPTOR.message_types_by_name['SourceProto'] = _SOURCEPROTO
DESCRIPTOR.message_types_by_name['TargetProto'] = _TARGETPROTO
DESCRIPTOR.message_types_by_name['FamilyFallbackProto'] = _FAMILYFALLBACKPROTO
DESCRIPTOR.message_types_by_name['RegionProto'] = _REGIONPROTO
DESCRIPTOR.message_types_by_name['ScriptProto'] = _SCRIPTPROTO
DESCRIPTOR.message_types_by_name['LanguageProto'] = _LANGUAGEPROTO
DESCRIPTOR.message_types_by_name['ExemplarCharsProto'] = _EXEMPLARCHARSPROTO
DESCRIPTOR.message_types_by_name['SampleTextProto'] = _SAMPLETEXTPROTO
DESCRIPTOR.message_types_by_name['GlyphGroupProto'] = _GLYPHGROUPPROTO
DESCRIPTOR.enum_types_by_name['TargetTypeProto'] = _TARGETTYPEPROTO
DESCRIPTOR.enum_types_by_name['FontType'] = _FONTTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FamilyProto = _reflection.GeneratedProtocolMessageType('FamilyProto', (_message.Message,), {

  'RegistryDefaultOverridesEntry' : _reflection.GeneratedProtocolMessageType('RegistryDefaultOverridesEntry', (_message.Message,), {
    'DESCRIPTOR' : _FAMILYPROTO_REGISTRYDEFAULTOVERRIDESENTRY,
    '__module__' : 'fonts_public_pb2'
    # @@protoc_insertion_point(class_scope:google.fonts_public.FamilyProto.RegistryDefaultOverridesEntry)
    })
  ,
  'DESCRIPTOR' : _FAMILYPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.FamilyProto)
  })
_sym_db.RegisterMessage(FamilyProto)
_sym_db.RegisterMessage(FamilyProto.RegistryDefaultOverridesEntry)

FontProto = _reflection.GeneratedProtocolMessageType('FontProto', (_message.Message,), {
  'DESCRIPTOR' : _FONTPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.FontProto)
  })
_sym_db.RegisterMessage(FontProto)

AxisSegmentProto = _reflection.GeneratedProtocolMessageType('AxisSegmentProto', (_message.Message,), {
  'DESCRIPTOR' : _AXISSEGMENTPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.AxisSegmentProto)
  })
_sym_db.RegisterMessage(AxisSegmentProto)

SourceProto = _reflection.GeneratedProtocolMessageType('SourceProto', (_message.Message,), {
  'DESCRIPTOR' : _SOURCEPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.SourceProto)
  })
_sym_db.RegisterMessage(SourceProto)

TargetProto = _reflection.GeneratedProtocolMessageType('TargetProto', (_message.Message,), {
  'DESCRIPTOR' : _TARGETPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.TargetProto)
  })
_sym_db.RegisterMessage(TargetProto)

FamilyFallbackProto = _reflection.GeneratedProtocolMessageType('FamilyFallbackProto', (_message.Message,), {
  'DESCRIPTOR' : _FAMILYFALLBACKPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.FamilyFallbackProto)
  })
_sym_db.RegisterMessage(FamilyFallbackProto)

RegionProto = _reflection.GeneratedProtocolMessageType('RegionProto', (_message.Message,), {
  'DESCRIPTOR' : _REGIONPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.RegionProto)
  })
_sym_db.RegisterMessage(RegionProto)

ScriptProto = _reflection.GeneratedProtocolMessageType('ScriptProto', (_message.Message,), {
  'DESCRIPTOR' : _SCRIPTPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.ScriptProto)
  })
_sym_db.RegisterMessage(ScriptProto)

LanguageProto = _reflection.GeneratedProtocolMessageType('LanguageProto', (_message.Message,), {
  'DESCRIPTOR' : _LANGUAGEPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.LanguageProto)
  })
_sym_db.RegisterMessage(LanguageProto)

ExemplarCharsProto = _reflection.GeneratedProtocolMessageType('ExemplarCharsProto', (_message.Message,), {
  'DESCRIPTOR' : _EXEMPLARCHARSPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.ExemplarCharsProto)
  })
_sym_db.RegisterMessage(ExemplarCharsProto)

SampleTextProto = _reflection.GeneratedProtocolMessageType('SampleTextProto', (_message.Message,), {
  'DESCRIPTOR' : _SAMPLETEXTPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.SampleTextProto)
  })
_sym_db.RegisterMessage(SampleTextProto)

GlyphGroupProto = _reflection.GeneratedProtocolMessageType('GlyphGroupProto', (_message.Message,), {
  'DESCRIPTOR' : _GLYPHGROUPPROTO,
  '__module__' : 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts_public.GlyphGroupProto)
  })
_sym_db.RegisterMessage(GlyphGroupProto)


DESCRIPTOR._options = None
_FAMILYPROTO_REGISTRYDEFAULTOVERRIDESENTRY._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
"""Exposes information of interest buried in the CLDR.

Specifically useful are language and region info.

See more at http://cldr.unicode.org/.
"""

from collections import defaultdict
from collections import deque
import os
import tempfile
from urllib import request
from lxml import etree
import zipfile

CLDR_ZIP_URL = 'http://unicode.org/Public/cldr/latest/core.zip'
SUPPLEMENTAL_DATA_XML_PATH = 'common/supplemental/supplementalData.xml'
SUPPLEMENTAL_METADATA_XML_PATH = 'common/supplemental/supplementalMetadata.xml'
LANG_XML_PATH = 'common/main/{lang_code}.xml'
EN = 'en'
WORLD_REGION = '001'
NO_COMMENTS_XML_PARSER = etree.XMLParser(remove_comments=True)


class Cldr():
  """Exposes information of interest buried in the CLDR.

  Specifically useful are language and region info.

  See more at http://cldr.unicode.org/.
  """

  def _GetLangXmlPath(lang_code):
    return LANG_XML_PATH.format(lang_code=lang_code)

  def __init__(self):
    self._zip_dir = tempfile.TemporaryDirectory()
    self._DownloadCldrZip()

    lang_aliases = self._ParseLanguageAliases()
    lang_names, region_names = self._ParseDisplayNamesFromEnXml()
    lang_names_supplement, region_names_supplement = self._ParseDisplayNamesFromSupplementalData()
    lang_names.update(lang_names_supplement)
    region_names.update(region_names_supplement)
    region_populations, lang_populations, lang_regions = self._ParseLanguageRegionInfo()
    region_groups = self._ParseRegionGroups()

    self.langs = self._CompileLanguages(lang_aliases, lang_names, lang_populations)
    self.regions = self._CompileRegions(region_names, region_populations)
    self.region_groups = self._CompileRegionGroups(region_groups, region_names)
    self.lang_region_map = lang_regions
    self._SetRegionsOnLanguages(lang_regions)
    self._SetExemplarCharsOnLanguages()

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    self._zip_dir.cleanup()

  def _DownloadCldrZip(self):
    with tempfile.NamedTemporaryFile(suffix='.zip') as zip_file:
      request.urlretrieve(CLDR_ZIP_URL, zip_file.name)
      with zipfile.ZipFile(zip_file.name, 'r') as zip_ref:
        zip_ref.extractall(self._zip_dir.name)

  def _ParseLanguageAliases(self):
    root = etree.parse(os.path.join(self._zip_dir.name,
                       SUPPLEMENTAL_METADATA_XML_PATH), parser=NO_COMMENTS_XML_PARSER)
    aliases_data = root.find('.//{*}alias').xpath('languageAlias')
    aliases = defaultdict(lambda: set())
    for alias_data in aliases_data:
      aliases[alias_data.get('replacement')].add(alias_data.get('type'))
    return aliases

  def _ParseDisplayNamesFromEnXml(self):
    root = etree.parse(os.path.join(self._zip_dir.name,
                       Cldr._GetLangXmlPath(EN)), parser=NO_COMMENTS_XML_PARSER)
    languages_data = root.find('.//{*}languages')
    regions_data = root.find('.//{*}territories')

    lang_names = {lang.get('type'): lang.text for lang in languages_data}
    region_names = {region.get('type'): region.text for region in regions_data}
    return lang_names, region_names

  def _ParseDisplayNamesFromSupplementalData(self):
    root = etree.parse(os.path.join(
        self._zip_dir.name, SUPPLEMENTAL_DATA_XML_PATH))
    regions_data = root.find('.//{*}territoryInfo')

    lang_names = {}
    region_names = {}
    for region_data in regions_data:
      region_code = region_data.get('type')
      region_names[region_code] = region_data[0].text
      assert len(region_data) % 2 == 1, 'Is there a node without a comment label?'
      for i in range(1, len(region_data), 2):
        lang_code = region_data[i].get('type')
        lang_names[lang_code] = region_data[i+1].text
    return lang_names, region_names

  def _ParseLanguageRegionInfo(self):
    root = etree.parse(os.path.join(self._zip_dir.name,
                       SUPPLEMENTAL_DATA_XML_PATH), parser=NO_COMMENTS_XML_PARSER)
    regions_data = root.find('.//{*}territoryInfo')

    region_populations = defaultdict(lambda: 0)
    lang_populations = defaultdict(lambda: 0)
    lang_regions = defaultdict(lambda: set())

    for region_data in regions_data:
      if not len(region_data):
        # Skip pseudo entries like 'ZZ' (Unknown Region)
        continue
      region_code = region_data.get('type')
      region_population = int(region_data.get('population'))
      region_populations[region_code] = region_population
      for lang_population_data in region_data:
        lang_code = lang_population_data.get('type')
        population_percent = float(
            lang_population_data.get('populationPercent'))
        lang_populations[lang_code] += region_population * \
            population_percent / 100
        lang_regions[lang_code].add(region_code)

    return region_populations, lang_populations, lang_regions

  def _ParseRegionGroups(self):
    root = etree.parse(os.path.join(self._zip_dir.name,
                       SUPPLEMENTAL_DATA_XML_PATH), parser=NO_COMMENTS_XML_PARSER)
    territory_containment_data = root.find('.//{*}territoryContainment')
    groups_data = territory_containment_data.xpath('group[not(@status)]')
    raw_groups = {group_data.get('type'): group_data.get(
        'contains').split(' ') for group_data in groups_data}

    region_groups = {}

    # Expand region groups
    for group in raw_groups[WORLD_REGION]:
      regions = deque([group])
      i = 0
      while i < len(regions):
        if regions[i] in raw_groups:
          regions.extend(raw_groups[regions[i]])
          del regions[i]
        else:
          i += 1
      region_groups[group] = set(regions)

    return region_groups

  def _ParseExemplarCharacters(self, lang_code):
    xml_file = os.path.join(self._zip_dir.name, Cldr._GetLangXmlPath(lang_code))
    if not os.path.isfile(xml_file):
      return {}
    root = etree.parse(xml_file, parser=NO_COMMENTS_XML_PARSER)
    characters_data = root.find('.//{*}characters')
    if characters_data is None:
      return {}

    exemplar_chars = {}
    for exemplar_characters_data in characters_data.xpath('exemplarCharacters'):
      category = exemplar_characters_data.get('type') or 'base'
      chars = exemplar_characters_data.text
      chars = chars[1:len(chars)-1].split(' ')
      exemplar_chars[category] = chars
    return exemplar_chars

  def _CompileLanguages(self, lang_aliases, lang_names, lang_populations):
    langs = {}
    for lang_code in lang_populations:
      population = lang_populations[lang_code]
      display_name = lang_names[lang_code]
      assert display_name is not None, 'Missing display name for language: ' + lang_code
      aliases = [] if lang_code not in lang_aliases else lang_aliases[lang_code]
      langs[lang_code] = self.Language(lang_code, aliases, display_name, population)
    return langs

  def _CompileRegions(self, region_names, region_populations):
    regions = {}
    for region_code in region_populations:
      population = region_populations[region_code]
      display_name = region_names[region_code]
      assert display_name is not None, 'Missing display name for region: ' + region_code
      regions[region_code] = self.Region(region_code, display_name, population)
    return regions

  def _CompileRegionGroups(self, region_groups, region_names):
    return {region_names[group]: region_groups[group] for group in region_groups}

  def _SetRegionsOnLanguages(self, lang_regions):
    for lang_code in lang_regions:
      assert lang_code in self.langs, 'Missing language: ' + lang_code
      regions = [self.regions[r] for r in lang_regions[lang_code]]
      self.langs[lang_code].SetRegions(regions)

  def _SetExemplarCharsOnLanguages(self):
    for lang_code in self.langs:
      exemplar_chars = self._ParseExemplarCharacters(lang_code)
      self.langs[lang_code].SetExemplarChars(exemplar_chars)

  class Language():

    def __init__(self, lang_code, aliases, display_name, population):
      self.id = lang_code
      self.aliases = aliases
      self.name = display_name
      self.population = int(population)
      self.regions = None
      self.exemplar_chars = None

    def SetRegions(self, regions):
      assert self.regions is None, 'Regions already set on language: ' + self.id
      self.regions = regions

    def SetExemplarChars(self, exemplar_chars):
      assert self.exemplar_chars is None, 'Exemplar chars already set on language: ' + self.id
      self.exemplar_chars = exemplar_chars

  class Region():

    def __init__(self, region_code, display_name, population):
      self.id = region_code
      self.name = display_name
      self.population = int(population)



"""
Functions to fix fonts so they conform to the Google Fonts
specification:
https://github.com/googlefonts/gf-docs/tree/main/Spec
"""
from fontTools.misc.fixedTools import otRound
from fontTools.ttLib import TTFont, newTable, getTableModule
from fontTools.ttLib.tables import ttProgram
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable
from fontTools.ttLib.tables._f_v_a_r import NamedInstance
from gftools.util.google_fonts import _KNOWN_WEIGHTS
from gftools.utils import (
    download_family_from_Google_Fonts,
    Google_Fonts_has_family,
    font_stylename,
    font_familyname,
    family_bounding_box,
    get_unencoded_glyphs,
    normalize_unicode_marks,
    partition_cmap,
    typo_metrics_enabled,
    validate_family,
    unique_name,
)
from gftools.stat import gen_stat_tables

from os.path import basename, splitext
from copy import deepcopy
import logging


log = logging.getLogger(__name__)


__all__ = [
    "remove_tables",
    "add_dummy_dsig",
    "fix_unhinted_font",
    "fix_hinted_font",
    "fix_fs_type",
    "fix_weight_class",
    "fix_fs_selection",
    "fix_mac_style",
    "fix_fvar_instances",
    "update_nametable",
    "fix_nametable",
    "inherit_vertical_metrics",
    "fix_vertical_metrics",
    "fix_ascii_fontmetadata",
    "drop_nonpid0_cmap",
    "drop_mac_cmap",
    "fix_pua",
    "fix_isFixedPitch",
    "drop_mac_names",
    "drop_superfluous_mac_names",
    "fix_font",
    "fix_family",
    "rename_font",
    "fix_filename"
]


# The _KNOWN_WEIGHT_VALUES constant is used internally by the GF Engineering
# team so we cannot update ourselves. TODO (Marc F) unify this one day
WEIGHT_NAMES = _KNOWN_WEIGHTS
del WEIGHT_NAMES[""]
WEIGHT_NAMES["Hairline"] = 1
WEIGHT_NAMES["ExtraBlack"] = 1000
WEIGHT_VALUES = {v: k for k, v in WEIGHT_NAMES.items()}


UNWANTED_TABLES = frozenset(
    [
        "FFTM",
        "TTFA",
        "TSI0",
        "TSI1",
        "TSI2",
        "TSI3",
        "TSI5",
        "prop",
        "MVAR",
        "Debg",
    ]
)


def remove_tables(ttFont, tables=None):
    """Remove unwanted tables from a font. The unwanted tables must belong
    to the UNWANTED_TABLES set.

    Args:
        ttFont: a TTFont instance
        tables: an iterable containing tables remove
    """
    tables_to_remove = UNWANTED_TABLES if not tables else frozenset(tables)
    font_tables = frozenset(ttFont.keys())

    tables_not_in_font = tables_to_remove - font_tables
    if tables_not_in_font:
        log.warning(
            f"Cannot remove tables '{list(tables_not_in_font)}' since they are "
            f"not in the font."
        )

    required_tables = tables_to_remove - UNWANTED_TABLES
    if required_tables:
        log.warning(
            f"Cannot remove tables '{list(required_tables)}' since they are required"
        )

    tables_to_remove = UNWANTED_TABLES & font_tables & tables_to_remove
    if not tables_to_remove:
        return
    log.info(f"Removing tables '{list(tables_to_remove)}' from font")
    for tbl in tables_to_remove:
        del ttFont[tbl]


def add_dummy_dsig(ttFont):
    """Add a dummy dsig table to a font. Older versions of MS Word
    require this table.

    Args:
        ttFont: a TTFont instance
    """
    newDSIG = newTable("DSIG")
    newDSIG.ulVersion = 1
    newDSIG.usFlag = 0
    newDSIG.usNumSigs = 0
    newDSIG.signatureRecords = []
    ttFont.tables["DSIG"] = newDSIG


def fix_unhinted_font(ttFont):
    """Improve the appearance of an unhinted font on Win platforms by:
        - Add a new GASP table with a newtable that has a single
          range which is set to smooth.
        - Add a new prep table which is optimized for unhinted fonts.

    Args:
        ttFont: a TTFont instance
    """
    gasp = newTable("gasp")
    # Set GASP so all sizes are smooth
    gasp.gaspRange = {0xFFFF: 15}

    program = ttProgram.Program()
    assembly = ["PUSHW[]", "511", "SCANCTRL[]", "PUSHB[]", "4", "SCANTYPE[]"]
    program.fromAssembly(assembly)

    prep = newTable("prep")
    prep.program = program

    ttFont["gasp"] = gasp
    ttFont["prep"] = prep


def fix_hinted_font(ttFont):
    """Improve the appearance of a hinted font on Win platforms by enabling
    the head table's flag 3.

    Args:
        ttFont: a TTFont instance
    """
    if not 'fpgm' in ttFont:
        return False, ["Skipping. Font is not hinted."]
    old = ttFont["head"].flags
    ttFont["head"].flags |= 1 << 3
    return ttFont["head"].flags != old


def fix_fs_type(ttFont):
    """Set the OS/2 table's fsType flag to 0 (Installable embedding).

    Args:
        ttFont: a TTFont instance
    """
    old = ttFont["OS/2"].fsType
    ttFont["OS/2"].fsType = 0
    return old != 0


def fix_weight_class(ttFont):
    """Set the OS/2 table's usWeightClass so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    old_weight_class = ttFont["OS/2"].usWeightClass

    if 'fvar' in ttFont:
        fvar = ttFont['fvar']
        default_axis_values = {a.axisTag: a.defaultValue for a in fvar.axes}
        v = default_axis_values.get('wght', None)

        if v is not None:
            ttFont["OS/2"].usWeightClass = int(v)
            return ttFont["OS/2"].usWeightClass != old_weight_class

    stylename = font_stylename(ttFont)
    tokens = stylename.split()
    # Order WEIGHT_NAMES so longest names are first
    for style in sorted(WEIGHT_NAMES, key=lambda k: len(k), reverse=True):
        if style in tokens:
            ttFont["OS/2"].usWeightClass = WEIGHT_NAMES[style]
            return ttFont["OS/2"].usWeightClass != old_weight_class

    if "Italic" in tokens:
        ttFont["OS/2"].usWeightClass = 400
        return ttFont["OS/2"].usWeightClass != old_weight_class
    raise ValueError(
        f"Cannot determine usWeightClass because font style, '{stylename}' "
        f"doesn't have a weight token which is in our known "
        f"weights, '{WEIGHT_NAMES.keys()}'"
    )


def fix_fs_selection(ttFont):
    """Fix the OS/2 table's fsSelection so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = set(stylename.split())
    old_selection = fs_selection = ttFont["OS/2"].fsSelection

    # turn off all bits except for bit 7 (USE_TYPO_METRICS)
    fs_selection &= 1 << 7

    if "Italic" in tokens:
        fs_selection |= 1 << 0
    if "Bold" in tokens:
        fs_selection |= 1 << 5
    # enable Regular bit for all other styles
    if not tokens & set(["Bold", "Italic"]):
        fs_selection |= 1 << 6
    ttFont["OS/2"].fsSelection = fs_selection
    return old_selection != fs_selection


def fix_mac_style(ttFont):
    """Fix the head table's macStyle so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = set(stylename.split())
    mac_style = 0
    if "Italic" in tokens:
        mac_style |= 1 << 1
    if "Bold" in tokens:
        mac_style |= 1 << 0
    ttFont["head"].macStyle = mac_style


def fix_fvar_instances(ttFont):
    """Replace a variable font's fvar instances with a set of new instances
    that conform to the Google Fonts instance spec:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#fvar-instances

    Args:
        ttFont: a TTFont instance
    """
    if "fvar" not in ttFont:
        raise ValueError("ttFont is not a variable font")

    fvar = ttFont["fvar"]
    default_axis_vals = {a.axisTag: a.defaultValue for a in fvar.axes}

    stylename = font_stylename(ttFont)
    is_italic = "Italic" in stylename
    is_roman_and_italic = any(a for a in ("slnt", "ital") if a in default_axis_vals)

    wght_axis = next((a for a in fvar.axes if a.axisTag == "wght"), None)
    wght_min = int(wght_axis.minValue)
    wght_max = int(wght_axis.maxValue)

    nametable = ttFont["name"]

    def gen_instances(is_italic):
        results = []
        for wght_val in range(wght_min, wght_max + 100, 100):
            name = (
                WEIGHT_VALUES[wght_val]
                if not is_italic
                else f"{WEIGHT_VALUES[wght_val]} Italic".strip()
            )
            name = name.replace("Regular Italic", "Italic")

            coordinates = deepcopy(default_axis_vals)
            coordinates["wght"] = wght_val

            inst = NamedInstance()
            inst.subfamilyNameID = nametable.addName(name)
            inst.coordinates = coordinates
            results.append(inst)
        return results

    instances = []
    if is_roman_and_italic:
        for bool_ in (False, True):
            instances += gen_instances(is_italic=bool_)
    elif is_italic:
        instances += gen_instances(is_italic=True)
    else:
        instances += gen_instances(is_italic=False)
    fvar.instances = instances


def update_nametable(ttFont, family_name=None, style_name=None):
    """Update a static font's name table. The updated name table will conform
    to the Google Fonts support styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    If a style_name includes tokens other than wght and ital, these tokens
    will be appended to the family name e.g

    Input:
    family_name="MyFont"
    style_name="SemiCondensed SemiBold"

    Output:
    familyName (nameID 1) = "MyFont SemiCondensed SemiBold
    subFamilyName (nameID 2) = "Regular"
    typo familyName (nameID 16) = "MyFont SemiCondensed"
    typo subFamilyName (nameID 17) = "SemiBold"

    Google Fonts has used this model for several years e.g
    https://fonts.google.com/?query=cabin

    Args:
        ttFont:
        family_name: New family name
        style_name: New style name
    """
    if "fvar" in ttFont:
        raise ValueError("Cannot update the nametable for a variable font")
    nametable = ttFont["name"]

    # Remove nametable records which are not Win US English
    # TODO this is too greedy. We should preserve multilingual
    # names in the future. Please note, this has always been an issue.
    platforms = set()
    for rec in nametable.names:
        platforms.add((rec.platformID, rec.platEncID, rec.langID))
    platforms_to_remove = platforms ^ set([(3, 1, 0x409)])
    if platforms_to_remove:
        log.warning(
            f"Removing records which are not Win US English, {list(platforms_to_remove)}"
        )
        for platformID, platEncID, langID in platforms_to_remove:
            nametable.removeNames(
                platformID=platformID, platEncID=platEncID, langID=langID
            )

    # Remove any name records which contain linebreaks
    contains_linebreaks = []
    for r in nametable.names:
        for char in ("\n", "\r"):
            if char in r.toUnicode():
                contains_linebreaks.append(r.nameID)
    for nameID in contains_linebreaks:
        nametable.removeNames(nameID)

    if not family_name:
        family_name = font_familyname(ttFont)

    if not style_name:
        style_name = font_stylename(ttFont)

    ribbi = ("Regular", "Bold", "Italic", "Bold Italic")
    tokens = family_name.split() + style_name.split()

    nameids = {
        1: " ".join(t for t in tokens if t not in ribbi),
        2: " ".join(t for t in tokens if t in ribbi) or "Regular",
        16: " ".join(t for t in tokens if t not in list(WEIGHT_NAMES) + ['Italic']),
        17: " ".join(t for t in tokens if t in list(WEIGHT_NAMES) + ['Italic']) or "Regular"
    }
    # Remove typo name if they match since they're redundant
    if nameids[16] == nameids[1]:
        del nameids[16]
    if nameids[17] == nameids[2]:
        del nameids[17]

    family_name = nameids.get(16) or nameids.get(1)
    style_name = nameids.get(17) or nameids.get(2)

    # create NameIDs 3, 4, 6
    nameids[4] = f"{family_name} {style_name}"
    nameids[6] = f"{family_name.replace(' ', '')}-{style_name.replace(' ', '')}"
    nameids[3] = unique_name(ttFont, nameids)

    # Pass through all records and replace occurences of the old family name
    # with the new family name
    current_family_name = font_familyname(ttFont)
    for record in nametable.names:
        string = record.toUnicode()
        if current_family_name in string:
            nametable.setName(
                string.replace(current_family_name, family_name),
                record.nameID,
                record.platformID,
                record.platEncID,
                record.langID,
            )

    # Remove previous typographic names
    for nameID in (16, 17):
        nametable.removeNames(nameID=nameID)

    # Update nametable with new names
    for nameID, string in nameids.items():
        nametable.setName(string, nameID, 3, 1, 0x409)


def fix_nametable(ttFont):
    """Fix a static font's name table so it conforms to the Google Fonts
    supported styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    if "fvar" in ttFont:
        from fontTools.varLib.instancer.names import updateNameTable
        dflt_axes = {a.axisTag: a.defaultValue for a in ttFont['fvar'].axes}
        updateNameTable(ttFont, dflt_axes)
        return
    family_name = font_familyname(ttFont)
    style_name = font_stylename(ttFont)
    update_nametable(ttFont, family_name, style_name)


def rename_font(font, new_name):
    nametable = font["name"]
    current_name = font_familyname(font)
    if not current_name:
        raise Exception(
            "Name table does not contain nameID 1 or nameID 16. "
            "This tool does not work on webfonts."
        )
    log.info("Updating font name records")
    for record in nametable.names:
        record_string = record.toUnicode()

        no_space = current_name.replace(" ", "")
        hyphenated = current_name.replace(" ", "-")
        if " " not in record_string:
            new_string = record_string.replace(no_space, new_name.replace(" ", ""))
        else:
            new_string = record_string.replace(current_name, new_name)

        if new_string is not record_string:
            record_info = (
                record.nameID,
                record.platformID,
                record.platEncID,
                record.langID
            )
            log.info(
                "Updating {}: '{}' to '{}'".format(
                    record_info,
                    record_string,
                    new_string,
                )
            )
            record.string = new_string


def fix_filename(ttFont):
    ext = splitext(ttFont.reader.file.name)[1]
    family_name = font_familyname(ttFont)
    style_name = font_stylename(ttFont)

    if "fvar" in ttFont:
        axes = ",".join([a.axisTag for a in ttFont['fvar'].axes])
        if "Italic" in style_name:
            return f"{family_name}-Italic[{axes}]{ext}".replace(" ", "")
        return f"{family_name}[{axes}]{ext}".replace(" ", "")
    return f"{family_name}-{style_name}{ext}".replace(" ", "")


def inherit_vertical_metrics(ttFonts, family_name=None):
    """Inherit the vertical metrics from the same family which is
    hosted on Google Fonts.

    Args:
        ttFonts: a list of TTFont instances which belong to a family
        family_name: Optional string which allows users to specify a
            different family to inherit from e.g "Maven Pro".
    """
    family_name = font_familyname(ttFonts[0]) if not family_name else family_name

    gf_fonts = list(map(TTFont, download_family_from_Google_Fonts(family_name)))
    gf_fonts = {font_stylename(f): f for f in gf_fonts}
    # TODO (Marc F) use Regular font instead. If VF use font which has Regular
    # instance
    gf_fallback = list(gf_fonts.values())[0]

    fonts = {font_stylename(f): f for f in ttFonts}
    for style, font in fonts.items():
        if style in gf_fonts:
            src_font = gf_fonts[style]
        else:
            src_font = gf_fallback
        copy_vertical_metrics(src_font, font)

        if typo_metrics_enabled(src_font):
            font["OS/2"].fsSelection |= 1 << 7


def fix_vertical_metrics(ttFonts):
    """Fix a family's vertical metrics based on:
    https://github.com/googlefonts/gf-docs/tree/main/VerticalMetrics

    Args:
        ttFonts: a list of TTFont instances which belong to a family
    """
    src_font = next((f for f in ttFonts if font_stylename(f) == "Regular"), ttFonts[0])

    # TODO (Marc F) CJK Fonts?

    # If OS/2.fsSelection bit 7 isn't enabled, enable it and set the typo metrics
    # to the previous win metrics.
    if not typo_metrics_enabled(src_font):
        src_font["OS/2"].fsSelection |= 1 << 7  # enable USE_TYPO_METRICS
        src_font["OS/2"].sTypoAscender = src_font["OS/2"].usWinAscent
        src_font["OS/2"].sTypoDescender = -src_font["OS/2"].usWinDescent
        src_font["OS/2"].sTypoLineGap = 0

    # Set the hhea metrics so they are the same as the typo
    src_font["hhea"].ascent = src_font["OS/2"].sTypoAscender
    src_font["hhea"].descent = src_font["OS/2"].sTypoDescender
    src_font["hhea"].lineGap = src_font["OS/2"].sTypoLineGap

    # Set the win Ascent and win Descent to match the family's bounding box
    win_desc, win_asc = family_bounding_box(ttFonts)
    src_font["OS/2"].usWinAscent = win_asc
    src_font["OS/2"].usWinDescent = abs(win_desc)

    # Set all fonts vertical metrics so they match the src_font
    for ttFont in ttFonts:
        ttFont["OS/2"].fsSelection |= 1 << 7
        copy_vertical_metrics(src_font, ttFont)


def copy_vertical_metrics(src_font, dst_font):
    for table, key in [
        ("OS/2", "usWinAscent"),
        ("OS/2", "usWinDescent"),
        ("OS/2", "sTypoAscender"),
        ("OS/2", "sTypoDescender"),
        ("OS/2", "sTypoLineGap"),
        ("hhea", "ascent"),
        ("hhea", "descent"),
        ("hhea", "lineGap"),
    ]:
        val = getattr(src_font[table], key)
        setattr(dst_font[table], key, val)


def fix_italic_angle(ttFont):
    style_name = font_stylename(ttFont)
    if "Italic" not in style_name and ttFont["post"].italicAngle != 0:
        ttFont["post"].italicAngle = 0
    # TODO (Marc F) implement for italic fonts


def fix_ascii_fontmetadata(font):
    """Fixes TTF 'name' table strings to be ascii only"""
    for name in font['name'].names:
        title = name.string.decode(name.getEncoding())
        title = normalize_unicode_marks(title)
        name.string = title.encode(name.getEncoding())


def convert_cmap_subtables_to_v4(font):
  """Converts all cmap subtables to format 4.

  Returns a list of tuples (format, platformID, platEncID) of the tables
  which needed conversion."""
  cmap = font['cmap']
  outtables = []
  converted = []
  for table in cmap.tables:
    if table.format != 4:
      converted.append((table.format, table.platformID, table.platEncID))
    newtable = CmapSubtable.newSubtable(4)
    newtable.platformID = table.platformID
    newtable.platEncID = table.platEncID
    newtable.language = table.language
    newtable.cmap = table.cmap
    outtables.append(newtable)
  font['cmap'].tables = outtables
  return converted


def drop_nonpid0_cmap(font, report=True):
  keep, drop = partition_cmap(font, lambda table: table.platformID == 0, report)
  return drop


def drop_mac_cmap(font, report=True):
  keep, drop = partition_cmap(font, lambda table: table.platformID != 1 or table.platEncID != 0, report)
  return drop

def fix_pua(font):
    unencoded_glyphs = get_unencoded_glyphs(font)
    if not unencoded_glyphs:
        return

    ucs2cmap = None
    cmap = font["cmap"]

    # Check if an UCS-2 cmap exists
    for ucs2cmapid in ((3, 1), (0, 3), (3, 0)):
        ucs2cmap = cmap.getcmap(ucs2cmapid[0], ucs2cmapid[1])
        if ucs2cmap:
            break
    # Create UCS-4 cmap and copy the contents of UCS-2 cmap
    # unless UCS 4 cmap already exists
    ucs4cmap = cmap.getcmap(3, 10)
    if not ucs4cmap:
        cmapModule = getTableModule('cmap')
        ucs4cmap = cmapModule.cmap_format_12(12)
        ucs4cmap.platformID = 3
        ucs4cmap.platEncID = 10
        ucs4cmap.language = 0
        if ucs2cmap:
            ucs4cmap.cmap = deepcopy(ucs2cmap.cmap)
        cmap.tables.append(ucs4cmap)
    # Map all glyphs to UCS-4 cmap Supplementary PUA-A codepoints
    # by 0xF0000 + glyphID
    ucs4cmap = cmap.getcmap(3, 10)
    for glyphID, glyph in enumerate(font.getGlyphOrder()):
        if glyph in unencoded_glyphs:
            ucs4cmap.cmap[0xF0000 + glyphID] = glyph
    font['cmap'] = cmap
    return True


def fix_isFixedPitch(ttfont):

    same_width = set()
    glyph_metrics = ttfont['hmtx'].metrics
    messages = []
    changed = False
    for character in [chr(c) for c in range(65, 91)]:
        same_width.add(glyph_metrics[character][0])

    if len(same_width) == 1:
        if ttfont['post'].isFixedPitch == 1:
            messages.append("Skipping isFixedPitch is set correctly")
        else:
            messages.append("Font is monospace. Updating isFixedPitch to 0")
            ttfont['post'].isFixedPitch = 1
            changed = True

        familyType = ttfont['OS/2'].panose.bFamilyType
        if familyType == 2:
            expected = 9
        elif familyType == 3 or familyType == 5:
            expected = 3
        elif familyType == 0:
            messages.append("Font is monospace but panose fields seems to be not set."
                  " Setting values to defaults (FamilyType = 2, Proportion = 9).")
            ttfont['OS/2'].panose.bFamilyType = 2
            ttfont['OS/2'].panose.bProportion = 9
            changed = True
            expected = None
        else:
            expected = None

        if expected:
            if ttfont['OS/2'].panose.bProportion == expected:
                messages.append("Skipping OS/2.panose.bProportion is set correctly")
            else:
                messages.append(("Font is monospace."
                       " Since OS/2.panose.bFamilyType is {}"
                       " we're updating OS/2.panose.bProportion"
                       " to {}").format(familyType, expected))
                ttfont['OS/2'].panose.bProportion = expected
                changed = True

        widths = [m[0] for m in ttfont['hmtx'].metrics.values() if m[0] > 0]
        width_max = max(widths)
        if ttfont['hhea'].advanceWidthMax == width_max:
            messages.append("Skipping hhea.advanceWidthMax is set correctly")
        else:
            messsages.append("Font is monospace. Updating hhea.advanceWidthMax to %i" %
                  width_max)
            ttfont['hhea'].advanceWidthMax = width_max
            changed = True

        avg_width = otRound(sum(widths) / len(widths))
        if avg_width == ttfont['OS/2'].xAvgCharWidth:
            messages.append("Skipping OS/2.xAvgCharWidth is set correctly")
        else:
            messages.append("Font is monospace. Updating OS/2.xAvgCharWidth to %i" %
                  avg_width)
            ttfont['OS/2'].xAvgCharWidth = avg_width
            changed = True
    else:
        if ttfont['post'].isFixedPitch != 0 or ttfont['OS/2'].panose.bProportion != 0:
            changed = True
        ttfont['post'].isFixedPitch = 0
        ttfont['OS/2'].panose.bProportion = 0
    return changed, messages


def drop_superfluous_mac_names(ttfont):
    """Drop superfluous Mac nameIDs.

    The following nameIDS are kept:
    1: Font Family name,
    2: Font Family Subfamily name,
    3: Unique font identifier,
    4: Full font name,
    5: Version string,
    6: Postscript name,
    16: Typographic family name,
    17: Typographic Subfamily name
    18: Compatible full (Macintosh only),
    20: PostScript CID,
    21: WWS Family Name,
    22: WWS Subfamily Name,
    25: Variations PostScript Name Prefix.

    We keep these IDs in order for certain application to still function
    such as Word 2011. IDs 1-6 are very common, > 16 are edge cases.

    https://www.microsoft.com/typography/otspec/name.htm"""
    keep_ids = [1, 2, 3, 4, 5, 6, 16, 17, 18, 20, 21, 22, 25]
    changed = False
    for n in range(255):
        if n not in keep_ids:
            name = ttfont['name'].getName(n, 1, 0, 0)
            if name:
                changed = True
                ttfont['name'].names.remove(name)
    return changed


def drop_mac_names(ttfont):
    """Drop all mac names"""
    changed = False
    for n in range(255):
        name = ttfont['name'].getName(n, 1, 0, 0)
        if name:
            ttfont['name'].names.remove(name)
            changed = True
    return changed


def fix_font(font, include_source_fixes=False, new_family_name=None):
    if new_family_name:
        rename_font(font, new_family_name)
    font["OS/2"].version = 4

    if "fpgm" in font:
        fix_hinted_font(font)
    else:
        fix_unhinted_font(font)

    if "fvar" in font:
        remove_tables(font, ["MVAR"])

    if include_source_fixes:
        log.warning(
            "include-source-fixes is enabled. Please consider fixing the "
            "source files instead."
        )
        remove_tables(font)
        fix_nametable(font)
        fix_fs_type(font)
        fix_fs_selection(font)
        fix_mac_style(font)
        fix_weight_class(font)
        fix_italic_angle(font)

        if "fvar" in font:
            fix_fvar_instances(font)


def fix_family(fonts, include_source_fixes=False, new_family_name=None):
    """Fix all fonts in a family"""
    validate_family(fonts)

    for font in fonts:
        fix_font(
            font,
            include_source_fixes=include_source_fixes,
            new_family_name=new_family_name
        )
    family_name = font_familyname(fonts[0])
    if include_source_fixes:
        try:
            if Google_Fonts_has_family(family_name):
                inherit_vertical_metrics(fonts)
            else:
                log.warning(
                    f"{family_name} is not on Google Fonts. Skipping "
                    "regression fixes"
                )
        except FileNotFoundError:
            log.warning(
                f"Google Fonts api key not found so we can't regression "
                "fix fonts. See Repo readme to add keys."
            )
        fix_vertical_metrics(fonts)
        if all(["fvar" in f for f in fonts]):
            gen_stat_tables(fonts, ["opsz", "wdth", "wght", "ital", "slnt"])


class FontFixer():
    def __init__(self, path, report=True, verbose=False, **kwargs):
        self.font = TTFont(path)
        self.path = path
        self.font_filename = basename(path)
        self.saveit = False
        self.report = report
        self.verbose = verbose
        self.messages = []
        self.args = kwargs
        self.fixes = []
        if "fixes" in kwargs:
            self.fixes = kwargs["fixes"]

    def __del__(self):
        if self.report:
            print("\n".join(self.messages))
        if self.saveit:
            if self.verbose:
                print('Saving %s to %s.fix' % (self.font_filename, self.path))
            self.font.save(self.path + ".fix")
        elif self.verbose:
            print('There were no changes needed on %s!' % self.font_filename)

    def show(self):
        pass

    def fix(self):
        for f in self.fixes:
            rv = f(self.font)
            if isinstance(rv, tuple) and len(rv) == 2:
                changed, messages = rv
                self.messages.extend(messages)
            else:
                changed = rv
            if changed:
                self.saveit = True


class GaspFixer(FontFixer):

    def fix(self, value=15):
        try:
            table = self.font.get('gasp')
            table.gaspRange[65535] = value
            self.saveit = True
        except:
            print(('ER: {}: no table gasp... '
                  'Creating new table. ').format(self.path))
            table = ttLib.newTable('gasp')
            table.gaspRange = {65535: value}
            self.font['gasp'] = table
            self.saveit = True

    def show(self):
        try:
            self.font.get('gasp')
        except:
            print('ER: {}: no table gasp'.format(self.path))
            return

        try:
            print(self.font.get('gasp').gaspRange[65535])
        except IndexError:
            print('ER: {}: no index 65535'.format(self.path))




#!/usr/bin/env python3
"""
gftools gen-html aka diffbrowsers2.

Generate html documents to proof a font family, or generate documents to
diff two families.

Examples:
# Generate proofing documents for a single font
gftools gen-html proof font1.ttf

# Generate proofing documents for a family of fonts
gftools gen-html proof font1.ttf font2.ttf font3.ttf

# Output test pages to a dir
gftools gen-html proof font1.ttf -o ~/Desktop/myFamily

# Generate proofing documents and output images using Browserstack
# (a subscription is required)
gftools gen-html proof font1.ttf --imgs

# Generate diff documents
gftools gen-html diff -fb ./fonts_before/font1.ttf -fa ./fonts_after/font1.ttf
"""
from pkg_resources import resource_filename
from gftools.html import HtmlProof, HtmlDiff
import argparse


def main():
    html_templates_dir = resource_filename("gftools", "templates")

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        dest="command", required=True, metavar='"proof" or "diff"'
    )

    # Optional args which can be used in all subparsers
    universal_options_parser = argparse.ArgumentParser(add_help=False)
    universal_options_parser.add_argument(
        "--pages",
        nargs="+",
        help="Choose which templates to populate. By default, all templates "
        "are populated.",
    )
    universal_options_parser.add_argument(
        "--pt-size", "-pt", help="Change pt size of document text", default=14
    )
    universal_options_parser.add_argument(
        "--imgs",
        action="store_true",
        help="Output images using Browserstack.",
    )
    universal_options_parser.add_argument(
        "--out", "-o", help="Output dir", default="diffbrowsers"
    )
    universal_options_parser.add_argument(
        "--template-dir",
        "-td",
        help="HTML template directory. By default, gftools/templates is used.",
        default=resource_filename("gftools", "templates"),
    )

    proof_parser = subparsers.add_parser(
        "proof",
        parents=[universal_options_parser],
        help="Generate html proofing documents for a family",
    )
    proof_parser.add_argument("fonts", nargs="+")

    diff_parser = subparsers.add_parser(
        "diff",
        parents=[universal_options_parser],
        help="Generate html diff documents which compares two families. "
        "Variable fonts can be compared against static fonts because we "
        "match the fvar instances against the static fonts. To Match fonts "
        "we use the font's name table records. For static fonts, the fullname "
        "is used e.g 'Maven Pro Medium'. For variable fonts, the family name "
        "+ fvar instance subfamilyname is used e.g 'Maven Pro' + 'Medium'.",
    )
    diff_parser.add_argument("--fonts-before", "-fb", nargs="+", required=True)
    diff_parser.add_argument("--fonts-after", "-fa", nargs="+", required=True)

    args = parser.parse_args()

    if args.command == "proof":
        html = HtmlProof(args.fonts, args.out, template_dir=args.template_dir)

    elif args.command == "diff":
        html = HtmlDiff(
            args.fonts_before,
            args.fonts_after,
            args.out,
            template_dir=args.template_dir,
        )

    html.build_pages(args.pages, pt_size=args.pt_size)

    if args.imgs:
        html.save_imgs()


if __name__ == "__main__":
    main()




STYLE_NAMES = ["Thin",
               "ExtraLight",
               "Light",
               "Regular",
               "Medium",
               "SemiBold",
               "Bold",
               "ExtraBold",
               "Black",
               "Thin Italic",
               "ExtraLight Italic",
               "Light Italic",
               "Italic",
               "Medium Italic",
               "SemiBold Italic",
               "Bold Italic",
               "ExtraBold Italic",
               "Black Italic"]

RIBBI_STYLE_NAMES = ["Regular",
                     "Italic",
                     "Bold",
                     "BoldItalic"]


def get_stylename(filename):
  filename_base = filename.split('.')[0]
  return filename_base.split('-')[-1]

def _familyname(filename):
  filename_base = filename.split('.')[0]
  names = filename_base.split('-')
  names.pop()
  return '-'.join(names)

def is_italic(stylename):
  return 'Italic' in stylename


def is_regular(stylename):
  return ("Regular" in stylename or
          (stylename in STYLE_NAMES and
           stylename not in RIBBI_STYLE_NAMES and
           "Italic" not in stylename))


def is_bold(stylename):
  return stylename in ["Bold", "BoldItalic"]


def is_filename_canonical(filename):
  if '-' not in filename:
    return False
  else:
    style = get_stylename(filename)
    for valid in STYLE_NAMES:
      valid = ''.join(valid.split(' '))
      if style == valid:
        return True
    # otherwise:
    return False



#!/usr/bin/env python3
"""
Rename a font.

Changes font menu name and filename. User can also specify their
own output path.

Usage:
gftools rename-font font.ttf "New Family Name"
"""
import argparse
from fontTools.ttLib import TTFont
from gftools.utils import font_familyname
from gftools.fix import rename_font


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("font")
    parser.add_argument("new_name")
    parser.add_argument("-o", "--out")
    args = parser.parse_args()

    font = TTFont(args.font)
    current_name = font_familyname(font)
    rename_font(font, args.new_name)

    if args.out:
        out = args.out
    else:
        out = args.font.replace(
            current_name.replace(" ", ""), args.new_name.replace(" ", "")
        )
    print("Saving font: {}".format(out))
    font.save(out)


if __name__ == "__main__":
    main()



import pytest
from glob import glob
from fontTools.ttLib import TTFont
import tempfile
import os
from gftools.html import *
import re
from copy import deepcopy


TEST_DATA = os.path.join("data", "test")


@pytest.fixture
def SimpleTemplate():
    class SimpleTemplate(HtmlTemplater):
        def __init__(self, out, template_dir):
            super().__init__(out=out, template_dir=template_dir)
    return SimpleTemplate


@pytest.fixture
def static_fonts():
    return [f for f in glob(os.path.join("data", "test", "mavenpro", "*.ttf"))]


@pytest.fixture
def static_ttfonts():
    return [TTFont(f) for f in glob(os.path.join("data", "test", "mavenpro", "*.ttf"))]


@pytest.fixture
def var_font():
    return os.path.join(TEST_DATA, "Inconsolata[wdth,wght].ttf")


@pytest.fixture
def var_ttfont():
    return TTFont(os.path.join(TEST_DATA, "Inconsolata[wdth,wght].ttf"))


@pytest.fixture
def var_font2():
    return os.path.join(TEST_DATA, "MavenPro[wght].ttf")


def _string_to_file(string, dst):
    with open(dst, "w") as doc:
        doc.write(string)
    return dst


def _file_to_string(fp):
    with open(fp) as f:
        return f.read()


def test_templating_basic(SimpleTemplate):
    with tempfile.TemporaryDirectory() as project_dir, tempfile.TemporaryDirectory() as template_dir:
        template_out = os.path.join(template_dir, "index.html")
        _string_to_file("<b>{{ text }}</b>", template_out)

        templater = SimpleTemplate(out=project_dir, template_dir=template_dir)
        templater.text = "Hello World"
        templater.build_pages(["index.html"])
        result = _file_to_string(templater.documents["index.html"].path)
        assert result == "<b>Hello World</b>"


def test_templating_with_multiple_variables(SimpleTemplate):
    with tempfile.TemporaryDirectory() as project_dir, tempfile.TemporaryDirectory() as template_dir:
        template_out = os.path.join(template_dir, "contact.html")
        _string_to_file("<b>{{ text }}</b><p>{{ url }}</p>", template_out)

        templater = SimpleTemplate(out=project_dir, template_dir=template_dir)
        templater.text = "Hello you"
        templater.url = "https://www.google.com"
        templater.build_pages(["contact.html"])
        result = _file_to_string(templater.documents["contact.html"].path)
        assert result == "<b>Hello you</b><p>https://www.google.com</p>"


def test_templating_populating_multiple_pages(SimpleTemplate):
    with tempfile.TemporaryDirectory() as project_dir, tempfile.TemporaryDirectory() as template_dir:
        template_out = os.path.join(template_dir, "index.html")
        _string_to_file("<b>{{ text }}</b>", template_out)
        template1_out = os.path.join(template_dir, "contact.html")
        _string_to_file("<b>{{ text }}</b><p>{{ url }}</p>", template1_out)

        templater = SimpleTemplate(out=project_dir, template_dir=template_dir)
        templater.text = "Hello World"
        templater.url = "https://www.google.com"
        templater.build_pages(["index.html", "contact.html"])

        assert len(templater.documents) == 2
        
        index_result = _file_to_string(templater.documents["index.html"].path)
        assert index_result == "<b>Hello World</b>"

        contact_result = _file_to_string(templater.documents["contact.html"].path)
        assert contact_result == "<b>Hello World</b><p>https://www.google.com</p>"


def test_CSSElement_used_as_fontface():
    font_face = CSSElement(
        selector="@font-face",
        font_family="Roboto",
        font_weight=400,
    )
    assert font_face.render() == "@font-face { font-family: Roboto; font-weight: 400; }"


def test_CSSEement_used_as_class():
    class_ = CSSElement(
        selector=".bold",
        font_family="Roboto",
        font_weight=700
    )
    assert class_.render() == ".bold { font-family: Roboto; font-weight: 700; }"


def test_CSSElement_private_attribs():
    class_with_private_variables = CSSElement(
        selector=".normal",
        _style="Foobar",
        font_weight=400
    )
    assert "Foobar" not in class_with_private_variables.render()


def test_css_font_weight(var_ttfont):
    var_ttfont["OS/2"].usWeightClass = 400
    assert css_font_weight(var_ttfont) == 400

    var_ttfont["OS/2"].usWeightClass = 250
    assert css_font_weight(var_ttfont) == 100


def _select_class(string, classes):
    return next((s for s in classes if string in s.selector), None)


def _select_font_face(string, classes):
    return next((s for s in classes if string in s.font_family), None)


def test_font_classes_from_static(static_ttfonts):
    css_classes = css_font_classes(static_ttfonts)
    assert len(css_classes) == len(static_ttfonts)

    regular = _select_class("Maven-Pro-Regular", css_classes)
    black = _select_class("Maven-Pro-Black", css_classes)

    assert regular.render() == (
        "Maven-Pro-Regular { font-family: Maven-Pro-Regular; font-weight: 400; "
        "font-style: normal; font-stretch: 100%; }"
    )
    assert black.render() == (
        "Maven-Pro-Black { font-family: Maven-Pro-Black; font-weight: 900; "
        "font-style: normal; font-stretch: 100%; }"
    )


def test_font_classes_from_vf(var_ttfont):
    css_classes = css_font_classes([var_ttfont])
    assert len(css_classes) == len(var_ttfont['fvar'].instances)

    semiexpanded_medium = _select_class("Inconsolata-SemiExpanded-Medium", css_classes)
    assert semiexpanded_medium.render() == (
        "Inconsolata-SemiExpanded-Medium { font-family: Inconsolata-Regular; "
        "font-weight: 500; font-style: normal; font-stretch: 112%; }"
    )
    extracondensed_black = _select_class("Inconsolata-ExtraCondensed-Black", css_classes)
    assert extracondensed_black.render() == (
        "Inconsolata-ExtraCondensed-Black { font-family: Inconsolata-Regular; "
        "font-weight: 900; font-style: normal; font-stretch: 62%; }"
    )


def test_font_faces_from_static(static_ttfonts):
    font_faces = css_font_faces(static_ttfonts)
    assert len(font_faces) == len(static_ttfonts)

    medium = _select_font_face("Maven-Pro-Medium", font_faces)
    assert medium.render() == (
        "@font-face { src: url(data/test/mavenpro/MavenPro-Medium.ttf); "
        "font-family: Maven-Pro-Medium; font-weight: 500; font-stretch: 100%; "
        "font-style: normal; }"
    )

    bold = _select_font_face("Maven-Pro-Bold", font_faces)
    assert bold.render() == (
        "@font-face { src: url(data/test/mavenpro/MavenPro-Bold.ttf); "
        "font-family: Maven-Pro-Bold; font-weight: 700; font-stretch: 100%; "
        "font-style: normal; }"
    )


def test_font_faces_from_vf(var_ttfont):
    font_faces = css_font_faces([var_ttfont])
    font_faces[0].render == (
        "@font-face { src: url(data/test/Inconsolata[wdth,wght].ttf); "
        "font-family: Inconsolata; font-weight: 200 900; font-stretch: 50% 200%; "
        "font-style: normal; }"
    )


def _font_faces_and_font_classes_linked(font_faces, css_classes):
    # font classes and font-faces are linked via the font-family property
    font_face_names = set(f.font_family for f in font_faces)
    css_class_names = set(c.font_family for c in css_classes)
    assert font_face_names == css_class_names


def test_font_faces_match_font_classes_static(static_ttfonts):
    font_faces = css_font_faces(static_ttfonts)
    css_classes = css_font_classes(static_ttfonts)
    _font_faces_and_font_classes_linked(font_faces, css_classes)


def test_font_faces_match_font_classes_vf(var_ttfont):
    font_faces = css_font_faces([var_ttfont])
    css_classes = css_font_classes([var_ttfont])
    _font_faces_and_font_classes_linked(font_faces, css_classes)


def _test_waterfall(waterfall_result, html):
    # Check pangram strings have been added.
    assert "quick wafting zephyrs vex bold jim." in waterfall_result

    font_faces = re.findall(r"@font-face", waterfall_result)
    assert len(font_faces) == len(html.css_font_faces)

    font_styles = re.findall(r".[A-z-]{1,50} \{ font-family", waterfall_result)
    assert len(font_styles) == len(html.css_font_classes)


def _test_text(text_result, html):
    assert "Resolution" in text_result


def test_HtmlProof_with_static_fonts(static_fonts):
    with tempfile.TemporaryDirectory() as project_dir:
        html = HtmlProof(
            out=project_dir,
            fonts=static_fonts
        )
        # Each font must have an @font-face and font_class
        assert len(html.fonts) == len(html.css_font_faces) == len(html.css_font_classes)
        html.build_pages(["waterfall.html", "text.html"])

        waterfall_result = _file_to_string(html.documents["waterfall.html"].path)
        _test_waterfall(waterfall_result, html)

        text_result = _file_to_string(html.documents["text.html"].path)
        _test_text(text_result, html)


def test_HtmlProof_with_vf(var_font):
    with tempfile.TemporaryDirectory() as project_dir:
        html = HtmlProof(
            out=project_dir,
            fonts=[var_font]
        )
        assert len(html.fonts) == len(html.css_font_faces)
        # VFs can contain multiple styles in a single font
        assert len(html.css_font_faces) < len(html.css_font_classes)
        html.build_pages(["waterfall.html"])

        waterfall_result = _file_to_string(html.documents["waterfall.html"].path)
        _test_waterfall(waterfall_result, html)


def test_HtmlDiff_match_css_classes_different_families(static_fonts, var_font):
    with tempfile.TemporaryDirectory() as project_dir:
        with pytest.raises(ValueError, match="No matching fonts found"):
            html = HtmlDiff(static_fonts, [var_font], project_dir)


def _check_css_classes_match(classes_before, classes_after):
    # Check css classes have same order
    before_properties = [
        (s.font_weight, s.font_stretch, s.font_style)
        for s in classes_before
    ]
    after_properties = [
        (s.font_weight, s.font_stretch, s.font_style)
        for s in classes_after
    ]
    assert before_properties == after_properties


def test_HtmlDiff_match_css_classes_different_styles(static_ttfonts):
    from gftools.fix import update_nametable
    family_before = static_ttfonts
    family_after = deepcopy(static_ttfonts)

    reg_after = next((f for f in family_after if "Regular.ttf" in f.reader.file.name), None)
    update_nametable(reg_after, style_name="Foobar")

    bold_after = next((f for f in family_after if "Bold.ttf" in f.reader.file.name), None)
    update_nametable(bold_after, style_name="Foobar2")

    with tempfile.TemporaryDirectory() as project_dir, tempfile.TemporaryDirectory() as mod_fonts:
        # save modified ttfonts to a tempdir and load them
        [f.save(os.path.join(mod_fonts, os.path.basename(f.reader.file.name))) for f in family_after]
        family_after = glob(os.path.join(mod_fonts, "*.ttf"))
        family_before = [f.reader.file.name for f in family_before]

        html = HtmlDiff(family_before, family_after, project_dir)
        # Check css classes do not contain
        for subfamily in ("Regular", "Bold", "Foobar", "Foobar2"):
            assert not any(subfamily in c._style for c in html.css_font_classes_before)
            assert not any(subfamily in c._style for c in html.css_font_classes_after)

        # Check css classes contain
        for subfamily in ("Medium", "Black"):
            assert any(subfamily in c._style for c in html.css_font_classes_before)
            assert any(subfamily in c._style for c in html.css_font_classes_after)

        _check_css_classes_match(html.css_font_classes_before, html.css_font_classes_after)


def test_HtmlDiff_match_css_classes_static_vs_vf(static_fonts, var_font2):
    # MavenPro VF has an fvar instance for each static font.
    # Every font in the static family should match an instance in Maven Pro VF
    with tempfile.TemporaryDirectory() as project_dir:
        html = HtmlDiff(static_fonts, [var_font2], project_dir)
        assert len(html.css_font_classes_before) == len(html.css_font_classes_after)
        _check_css_classes_match(html.css_font_classes_before, html.css_font_classes_after)




from fontFeatures import FontFeatures, Routine, Substitution
from fontFeatures.feaLib import FeaParser
from pathlib import Path
import logging

logger = logging.getLogger("ufomerge")
logging.basicConfig(level=logging.INFO)


def merge_ufos(
    ufo1,
    ufo2,
    glyphs=None,
    exclude_glyphs=None,
    codepoints=None,
    layout_handling="subset",
    existing_handling="replace",
):
    if glyphs is None:
        glyphs = []
    if exclude_glyphs:
        glyphs = set(glyphs) - set(exclude_glyphs)

    glyphs = set(glyphs)

    if codepoints:
        cp2glyph = {}
        for g in ufo2:
            for u in g.unicodes:
                cp2glyph[u] = g.name
        glyphs |= set(cp2glyph[c] for c in codepoints if c in cp2glyph)

    # Check those glyphs actually are in UFO 2
    not_there = glyphs - set(ufo2.keys())
    if len(not_there):
        logger.warn("The following glyphs were not in UFO 2: %s" % ", ".join(not_there))
        glyphs = glyphs - not_there

    if not glyphs:
        logger.info("No glyphs selected, nothing to do")
        exit(0)

    newglyphset = set(ufo1.keys()) | set(glyphs)

    # Handle layout subsetting here, in case closure is needed
    new_layout_rules = FontFeatures()
    if layout_handling == "ignore":
        pass
    else:
        includeDir = None
        if hasattr(ufo2, "_path"):
            includeDir = Path(ufo2._path).parent
        ff = FeaParser(ufo2.features.text, includeDir=includeDir).parse()
        for routine in ff.routines:
            newroutine = Routine(name=routine.name, flags=routine.flags)
            for rule in routine.rules:
                if not isinstance(rule, Substitution):
                    continue
                flat_outputs = [
                    item for sublist in rule.replacement for item in sublist
                ]
                rule.input = [list(set(r) & newglyphset) for r in rule.input]
                rule.precontext = [list(set(r) & newglyphset) for r in rule.precontext]
                rule.postcontext = [
                    list(set(r) & newglyphset) for r in rule.postcontext
                ]
                if (
                    any(not g for g in rule.input)
                    or any(not g for g in rule.precontext)
                    or any(not g for g in rule.postcontext)
                ):
                    continue
                if layout_handling == "closure":
                    # Any glyphs from "glyphs" substituted or generated by rules need to be added to the glyph set
                    if not any(g in glyphs for g in rule.involved_glyphs):
                        continue
                    glyphs |= set(flat_outputs)
                else:
                    # Any rules with new glyphs on the right hand side and glyphs
                    # we have on the left hand side need to be copied into UFO1
                    if not any(g in glyphs for g in flat_outputs):
                        continue
                    logging.debug("Adding rule '%s'", rule.asFea())
                newroutine.rules.append(rule)
            if newroutine.rules:
                # Was it in a feature?
                add_to = []
                for feature_name, routines in ff.features.items():
                    for routine_ref in routines:
                        if routine_ref.routine == routine:
                            add_to.append(feature_name)
                for feature_name in add_to:
                    new_layout_rules.addFeature(feature_name, [newroutine])

    # Kerning!!
    # # Create a list of flat kerning pairs for UFO 1
    # ufo1_kerns = set()
    # for l,r in ufo1.kerning.keys():
    #     l = ufo1.groups.get(l,[l])
    #     r = ufo1.groups.get(r,[r])
    #     for lg in l:
    #         for rg in r:
    #             ufo1_kerns.add((lg,rg))

    # Slim down the groups to only those in the glyph set
    for g in ufo2.groups.keys():
        ufo2.groups[g] = [g for g in ufo2.groups[g] if g in newglyphset]

    for (l, r), value in ufo2.kerning.items():
        lg = ufo2.groups.get(l, [l])
        rg = ufo2.groups.get(r, [r])
        if not lg or not rg:
            continue
        if any(lglyph not in newglyphset for lglyph in lg) or any(
            rglyph not in newglyphset for rglyph in rg
        ):
            continue
        # Just add for now. We should get fancy later
        ufo1.kerning[(l, r)] = value
        if l.startswith("public.kern"):
            if l not in ufo1.groups:
                ufo1.groups[l] = ufo2.groups[l]
            else:
                ufo1.groups[l] = list(set(ufo1.groups[l] + ufo2.groups[l]))
        if r.startswith("public.kern"):
            if r not in ufo1.groups:
                ufo1.groups[r] = ufo2.groups[r]
            else:
                ufo1.groups[r] = list(set(ufo1.groups[r] + ufo2.groups[r]))

    # Routines for merging font lib keys
    def merge_set(ufo1, ufo2, name, g, create_if_not_in_ufo1=False):
        if name not in ufo2.lib or g not in ufo2.lib[name]:
            return
        if name not in ufo1.lib:
            if create_if_not_in_ufo1:
                ufo1.lib[name] = []
            else:
                return
        if g not in ufo1.lib[name]:
            ufo1.lib[name].append(g)

    def merge_dict(ufo1, ufo2, name, g, create_if_not_in_ufo1=False):
        if name not in ufo2.lib or g not in ufo2.lib[name]:
            return
        if name not in ufo1.lib:
            if create_if_not_in_ufo1:
                ufo1.lib[name] = {}
            else:
                return
        ufo1.lib[name][g] = ufo2.lib[name][g]

    # Check the glyphs for components
    def close_components(glyphs, g):
        if not ufo2[g].components:
            return
        for comp in ufo2[g].components:
            if comp.baseGlyph not in newglyphset:
                # Well, this is the easy case
                glyphs.add(comp.baseGlyph)
                close_components(glyphs, comp.baseGlyph)
            elif existing_handling == "replace":
                # Also not a problem
                glyphs.add(comp.baseGlyph)
                close_components(glyphs, comp.baseGlyph)
            elif comp.baseGlyph in ufo1:
                # Oh bother.
                logger.warning(
                    f"New glyph {g} used component {comp.baseGlyph} which already exists in font; not replacing it, as you have not specified --replace-existing"
                )

    for g in list(glyphs):  # list() avoids "Set changed size during iteration" error
        close_components(glyphs, g)

    # Now do the add
    for g in glyphs:
        if existing_handling == "skip" and g in ufo1:
            logger.info("Skipping glyph '%s' already present in target file" % g)
            continue

        merge_set(ufo1, ufo2, "public.glyphOrder", g, create_if_not_in_ufo1=False)
        merge_set(ufo1, ufo2, "public.skipExportGlyphs", g, create_if_not_in_ufo1=True)
        merge_dict(ufo1, ufo2, "public.postscriptNames", g, create_if_not_in_ufo1=True)
        merge_dict(
            ufo1, ufo2, "public.openTypeCategories", g, create_if_not_in_ufo1=True
        )

        if g in ufo1:
            ufo1[g] = ufo2[g]
        else:
            ufo1.addGlyph(ufo2[g])

    if new_layout_rules.routines:
        ufo1.features.text += new_layout_rules.asFea(do_gdef=False)



from fontTools.ttLib import TTFont
import browserstack_screenshots
from pkg_resources import resource_filename
from jinja2 import Environment, FileSystemLoader, select_autoescape
from browserstack.local import Local
import tempfile
import os
from multiprocessing import Process
from contextlib import contextmanager
from http.server import *
import logging
import time
from copy import copy
import shutil
from collections import namedtuple
from gftools.utils import (
    font_sample_text,
    download_file,
    gen_gifs,
    font_familyname,
    font_stylename,
    font_is_italic,
    partition,
    get_encoded_glyphs,
)


__all__ = [
    "CSSElement",
    "css_font_class_from_static",
    "css_font_classes_from_vf",
    "css_font_faces",
    "css_font_classes",
    "HtmlTemplater",
    "HtmlProof",
    "HtmlDiff",
    "simple_server",
    "daemon_server",
    "browserstack_local",
    "css_font_weight",
]


log = logging.getLogger("gftools.html")
log.setLevel(logging.INFO)


WIDTH_CLASS_TO_CSS = {
    1: "50%",
    2: "62.5%",
    3: "75%",
    4: "87.5%",
    5: "100%",
    6: "112.5%",
    7: "125%",
    8: "150%",
    9: "200%",
}


# Browserstack configurations
# Browser list: https://www.browserstack.com/screenshots/browsers

# Latest working browsers offered by Browserstack. All of them support
# variable fonts.
# Note: As of (2020-12-15), these browsers are very
# old and differ when compared to their latest releases. Chrome
# 71 uses FreeType for VFs, this means that the hhea metrics
# will be used instead of the typo/win.
BSTACK_CONFIG_LATEST_BROWSERS = {
    "url": None,
    "local": True,
    "browsers": [
        {
            "os": "Windows",
            "browser_version": "71.0",
            "browser": "chrome",
            "os_version": "10",
        },
        # Catalina 13.0 and Mohave 12.0 don't produce a full size screenshot
        {
            "os": "OS X",
            "os_version": "High Sierra",
            "browser": "safari",
            "browser_version": "11.1",
        },
        # Browserstack doesn't have a version of Firefox that works correctly with
        # VFs. Support for VFs arrived in v62, however v62 and v63 on
        # Browserstack don't work.
        {
            "os": "Windows",
            "os_version": "10",
            "browser": "edge",
            "browser_version": "18.0",
        },
    ],
}

# Old browsers which use GDI ClearType. Only test on these browsers
# for more popular families (weekly views > 1 billion).
# Note: These browsers do not support variable fonts. As a safeguard,
# an exception will be raised if users attempt to use these browsers
# with variable fonts.
BSTACK_CONFIG_GDI_BROWSERS = {
    "url": None,
    "local": True,
    "browsers": [
        {
            "os": "Windows",
            "browser_version": "34.0",
            "browser": "chrome",
            "os_version": "7",
        },
        {
            "os": "Windows",
            "os_version": "7",
            "browser": "firefox",
            "browser_version": "50.0",
        },
        # Versions of IE on Win7 all use DW which look ok for unhinted
        # fonts.
    ],
}


class CSSElement(object):
    """Create a CSSElement. CSSElements include a render method which
    renders the class as a string so it can be used in html templates.

    Args:
      selector: The css selector e.g h1, h2, class-name, @font0face
      **kwargs: css properties and their property values e.g
        font_family="MyFamily"

    Example:
      | >>> bold = CSSElement("bold", font_weight=700, font_style="normal")
      | >>> bold.render()
      | >>> 'bold { font-weight: 700; font-style: normal; }'
    """

    def __init__(self, selector, **kwargs):
        self.selector = selector
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.declerations = {k.replace("_", "-"): v for k, v in kwargs.items()}

    def render(self):
        decleration_strings = " ".join(
            f"{k}: {v};" for k, v in self.declerations.items() if not k.startswith("-")
        )
        return f"{self.selector} {{ { decleration_strings } }}"


def css_font_faces(ttFonts, server_dir=None, position=None):
    """Generate @font-face CSSElements for a collection of fonts

    Args:
      ttFonts: a list containing ttFont instances
      server_dir: optional. A path to the root directory of the server.
        @font-face src urls are relative to the server's root dir.
      position: optional. Adds a suffix to the font-family name

    Returns:
      A list of @font-face CSSElements
    """
    results = []
    for ttFont in ttFonts:
        family_name = font_familyname(ttFont)
        style_name = font_stylename(ttFont)
        font_path = ttFont.reader.file.name
        path = (
            font_path
            if not server_dir
            else os.path.relpath(font_path, start=server_dir)
        )
        src = f"url({path})"
        font_family = _class_name(family_name, style_name, position)
        font_style = "italic" if font_is_italic(ttFont) else "normal"
        font_weight = css_font_weight(ttFont)
        font_stretch = WIDTH_CLASS_TO_CSS[ttFont["OS/2"].usWidthClass]

        if "fvar" in ttFont:
            fvar = ttFont["fvar"]
            axes = {a.axisTag: a for a in fvar.axes}
            if "wght" in axes:
                min_weight = int(axes["wght"].minValue)
                max_weight = int(axes["wght"].maxValue)
                font_weight = f"{min_weight} {max_weight}"
            if "wdth" in axes:
                min_width = int(axes["wdth"].minValue)
                max_width = int(axes["wdth"].maxValue)
                font_stretch = f"{min_width}% {max_width}%"
            if "ital" in axes:
                pass
            if "slnt" in axes:
                min_angle = int(axes["slnt"].minValue)
                max_angle = int(axes["slnt"].maxValue)
                font_style = f"oblique {min_angle}deg {max_angle}deg"

        font_face = CSSElement(
            "@font-face",
            src=src,
            font_family=font_family,
            font_weight=font_weight,
            font_stretch=font_stretch,
            font_style=font_style,
        )
        results.append(font_face)
    return results


def css_font_classes(ttFonts, position=None):
    """Generate class CSSElements for a collection of fonts

    Args:
      ttFonts: a list containing ttFont instances
      position: optional. Adds a suffix to the font-family name

    Returns:
      A list of class CSSElements
    """
    results = []
    for ttFont in ttFonts:
        if "fvar" in ttFont:
            results += css_font_classes_from_vf(ttFont, position)
        else:
            results.append(css_font_class_from_static(ttFont, position))
    return results


def _class_name(family_name, style_name, position=None):
    string = f"{family_name}-{style_name}".replace(" ", "-")
    return string if not position else f"{string}-{position}"


def css_font_weight(ttFont):
    # At Google Fonts, we released many Thin families with a weight class of
    # 250. This was implemented to fix older GDI browsers
    weight = ttFont["OS/2"].usWeightClass
    return weight if weight != 250 else 100


def css_font_class_from_static(ttFont, position=None):
    family_name = font_familyname(ttFont)
    style_name = font_stylename(ttFont)

    class_name = _class_name(family_name, style_name, position)
    font_family = class_name
    font_weight = css_font_weight(ttFont)
    font_style = "italic" if font_is_italic(ttFont) else "normal"
    font_stretch = WIDTH_CLASS_TO_CSS[ttFont["OS/2"].usWidthClass]
    return CSSElement(
        class_name,
        _full_name=f"{family_name} {style_name}",
        _style=style_name,
        _font_path=ttFont.reader.file.name,
        font_family=font_family,
        font_weight=font_weight,
        font_style=font_style,
        font_stretch=font_stretch,
    )


def css_font_classes_from_vf(ttFont, position=None):
    instances = ttFont["fvar"].instances
    nametable = ttFont["name"]
    family_name = font_familyname(ttFont)
    style_name = font_stylename(ttFont)

    results = []
    for instance in instances:
        nameid = instance.subfamilyNameID
        inst_style = nametable.getName(nameid, 3, 1, 0x409).toUnicode()

        class_name = _class_name(family_name, inst_style, position)
        font_family = _class_name(family_name, style_name, position)
        font_weight = (
            css_font_weight(ttFont)
            if not "wght" in instance.coordinates
            else int(instance.coordinates["wght"])
        )
        font_style = "italic" if "Italic" in inst_style else "normal"
        font_stretch = (
            "100%"
            if not "wdth" in instance.coordinates
            else f"{int(instance.coordinates['wdth'])}%"
        )
        font_class = CSSElement(
            class_name,
            _full_name=f"{family_name} {inst_style}",
            _style=inst_style,
            _font_path=ttFont.reader.file.name,
            font_family=font_family,
            font_weight=font_weight,
            font_style=font_style,
            font_stretch=font_stretch,
        )
        results.append(font_class)
    return results


Document = namedtuple("Document", ["filename", "path", "options"])


class HtmlTemplater(object):

    BROWSERSTACK_CONFIG = BSTACK_CONFIG_LATEST_BROWSERS

    def __init__(
        self,
        out="out",
        template_dir=resource_filename("gftools", "templates"),
        browserstack_username=None,
        browserstack_access_key=None,
        browserstack_config=None,
    ):
        """
        Generate html documents from Jinja2 templates and optionally
        screenshot the results on different browsers, using the
        Browserstack Screenshot api.

        When saving images, two brackground processes are started. A local
        server which serves the populated html documents
        and browserstack local. This allows Browserstack to take local
        screenshots.

        The main purpose of this class is to allow developers to
        write their own template generators by using inheritance e.g

        ```
        class MyTemplate(HtmlTemplater):
            def __init__(self, forename, surname, out):
                super().__init__(self, out)
                self.forename = forename
                self.surname = surname

        html = MyTemplate("Joe", "Doe")
        html.build_pages()
        ```

        template:
        <p>Hello {{ forename }} {{ surname }}.</p>

        result:
        <p>Hello Joe Doe.</p>

        For more complex examples, see HtmlProof and HtmlDiff in this
        module.

        All html docs and assets are saved into the specified out
        directory. Packaging the assets together makes it easier to share
        and we don't have to worry about absolute vs relative paths. This
        can be problematic for some assets such as webfonts where the path
        must be related to the local server, not the user's system.

        Note: Templates whose filename's start with an "_" are non
        renderable. This functionality exists so css classes etc
        can be defined in a _base.html file which other templates can
        inherit from.

        Args:
          out: output dir for generated html documents
          template_dir: the directory containing the html templates
          browserstack_username: optional. Browserstack username
          browserstack_access_key: optional. Browserstack access key
          browserstack_config: optional. Browserstack config file. See
            api docs for more info:
            https://www.browserstack.com/screenshots/api
        """
        self.template_dir = template_dir
        self.templates = [
            f
            for f in os.listdir(self.template_dir)
            if all([not f.startswith("_"), f.endswith("html")])
        ]
        # TODO we may want to make this an arg
        self.server_url = "http://0.0.0.0:8000"
        self.jinja = Environment(
            loader=FileSystemLoader(self.template_dir),
            autoescape=select_autoescape(["html", "xml"]),
        )

        self.out = self.mkdir(out)
        self.documents = {}
        # Set to true if html pages are going get cropped in Browserstack
        self.too_big_for_browserstack = False

        self.has_browserstack = (
            True
            if any([browserstack_access_key, "BSTACK_ACCESS_KEY" in os.environ])
            else False
        )

        if self.has_browserstack:
            self.browserstack_username = (
                browserstack_username or os.environ["BSTACK_USERNAME"]
            )
            self.browserstack_access_key = (
                browserstack_access_key or os.environ["BSTACK_ACCESS_KEY"]
            )
            self.browserstack_config = (
                browserstack_config if browserstack_config else self.BROWSERSTACK_CONFIG
            )
            auth = (self.browserstack_username, self.browserstack_access_key)
            self.screenshot = ScreenShot(auth=auth, config=self.browserstack_config)
        else:
            log.warning("No Browserstack credentials found. Image output disabled")

    def build_pages(self, pages=None, dst=None, **kwargs):
        if not pages:
            pages = self.templates
        log.info(f"Building pages {pages}")
        for page in pages:
            self.build_page(page, dst=dst, **kwargs)

    def build_page(self, filename, dst=None, **kwargs):
        if filename not in self.templates:
            raise ValueError(f"'{filename}' not in dir '{self.templates}'")
        if not "pt_size" in kwargs:
            kwargs["pt_size"] = 14
        # Combine self.__dict__ attributes with function kwargs. This allows Jinja
        # templates to access the class attributes
        jinja_kwargs = {**self.__dict__, **kwargs}
        out = self._render_html(filename, dst=dst, **jinja_kwargs)
        self.documents[filename] = Document(filename, out, kwargs)

    def _render_html(self, filename, dst=None, **kwargs):
        html_template = self.jinja.get_template(filename)
        html = html_template.render(**kwargs)
        dst = dst if dst else os.path.join(self.out, filename)
        with open(dst, "w") as html_file:
            html_file.write(html)
        return dst

    def mkdir(self, path):
        if not os.path.isdir(path):
            os.mkdir(path)
        return path

    def copy_files(self, srcs, dst):
        [shutil.copy(f, dst) for f in srcs]
        return [os.path.join(dst, os.path.basename(f)) for f in srcs]

    def save_imgs(self, pages=None, dst=None):
        assert hasattr(self, "screenshot")

        if self.too_big_for_browserstack:
            log.info("Images will be too big for Browserstack. Partitioning pages")
            self.partition()
        else:
            pages = pages if pages else self.documents.keys()
            log.warning("Generating images with Browserstack. This may take a while")
            img_dir = dst if dst else self.mkdir(os.path.join(self.out, "img"))
            with browserstack_local(self.browserstack_access_key), daemon_server(
                directory=self.out
            ):
                for page in pages:
                    if page not in self.documents:
                        raise ValueError(
                            f"{page} doesn't exist in documents, '{self.documents}'"
                        )
                    paths = self.documents[page].path
                    out = os.path.join(img_dir, page)
                    self.mkdir(out)
                    self._save_img(paths, out)

    def _save_img(self, path, dst):
        page = os.path.relpath(path, start=self.out)
        url = f"{self.server_url}/{page}"
        self.screenshot.take(url, dst)

    def partition(self):
        # Browserstack limits screenshots so they are less than 10000px tall.
        # To work around this limitation, we can patition the class into smaller
        # subclasses and then screenshot these instead
        raise NotImplementedError


class HtmlProof(HtmlTemplater):
    def __init__(
        self, fonts, out="out", template_dir=resource_filename("gftools", "templates")
    ):
        """Proof a single family."""
        super().__init__(out, template_dir=template_dir)
        fonts_dir = os.path.join(out, "fonts")
        self.mkdir(fonts_dir)

        self.fonts = self.copy_files(fonts, fonts_dir)
        self.ttFonts = [TTFont(f) for f in self.fonts]

        self.css_font_faces = css_font_faces(self.ttFonts, self.out)
        self.css_font_classes = css_font_classes(self.ttFonts)

        self.too_big_for_browserstack = len(self.css_font_classes) > 4

        self.sample_text = " ".join(font_sample_text(self.ttFonts[0]))
        # TODO to collect unencoded glyphs, we need to make a better version
        # of hbinput
        self.glyphs = get_encoded_glyphs(self.ttFonts[0])

    def partition(self):
        with tempfile.TemporaryDirectory() as tmp_out:
            temp_html = HtmlProof(
                fonts=self.fonts, out=tmp_out, template_dir=self.template_dir
            )
            # disable too_big_for_browserstack for subclasses otherwise we'll infinte loop
            temp_html.too_big_for_browserstack = False
            css_class_groups = partition(self.css_font_classes, 4)
            for idx, group in enumerate(css_class_groups):
                temp_html.css_font_classes = group
                for page, doc in self.documents.items():
                    temp_html.build_page(page, **doc.options)
                temp_html.save_imgs()
                src_imgs = os.path.join(temp_html.out, "img")
                dir_name = "-".join(s._style.replace(" ", "") for s in group)
                dst_imgs = os.path.join(self.out, "img", dir_name)
                shutil.rmtree(dst_imgs, ignore_errors=True)
                shutil.copytree(src_imgs, dst_imgs)


class HtmlDiff(HtmlTemplater):
    def __init__(
        self,
        fonts_before,
        fonts_after,
        out="out",
        template_dir=resource_filename("gftools", "templates"),
    ):
        """Compare two families"""
        super().__init__(out=out, template_dir=template_dir)
        fonts_before_dir = os.path.join(out, "fonts_before")
        fonts_after_dir = os.path.join(out, "fonts_after")
        self.mkdir(fonts_before_dir)
        self.mkdir(fonts_after_dir)

        self.fonts_before = self.copy_files(fonts_before, fonts_before_dir)
        self.ttFonts_before = [TTFont(f) for f in self.fonts_before]

        self.fonts_after = self.copy_files(fonts_after, fonts_after_dir)
        self.ttFonts_after = [TTFont(f) for f in self.fonts_after]

        self.css_font_faces_before = css_font_faces(
            self.ttFonts_before, self.out, position="before"
        )
        self.css_font_faces_after = css_font_faces(
            self.ttFonts_after, self.out, position="after"
        )

        self.css_font_classes_before = css_font_classes(self.ttFonts_before, "before")
        self.css_font_classes_after = css_font_classes(self.ttFonts_after, "after")
        self._match_css_font_classes()

        self.too_big_for_browserstack = len(self.css_font_classes_before) > 4

        self.sample_text = " ".join(font_sample_text(self.ttFonts_before[0]))
        self.glyphs = get_encoded_glyphs(self.ttFonts_before[0])

    def _match_css_font_classes(self):
        """Match css font classes by full names for static fonts and
        family name + instance name for fvar instances"""
        styles_before = {s._full_name: s for s in self.css_font_classes_before}
        styles_after = {s._full_name: s for s in self.css_font_classes_after}
        shared_styles = set(styles_before) & set(styles_after)

        self.css_font_classes_before = sorted(
            [s for k, s in styles_before.items() if k in shared_styles],
            key=lambda s: (s.font_weight, s._full_name),
        )
        self.css_font_classes_after = sorted(
            [s for k, s in styles_after.items() if k in shared_styles],
            key=lambda s: (s.font_weight, s._full_name),
        )
        if not all([self.css_font_classes_before, self.css_font_classes_after]):
            raise ValueError("No matching fonts found")

    def _render_html(
        self,
        filename,
        **kwargs,
    ):
        html_template = self.jinja.get_template(filename)

        # This document is intended for humans. It includes a button
        # to toggle which set of fonts is visible.
        combined = html_template.render(include_ui=True, **kwargs)
        combined_path = os.path.join(self.out, filename)
        with open(combined_path, "w") as combined_html:
            combined_html.write(combined)

        # This document contains fonts_before. It solely exists for
        # screenshot generation purposes
        before_kwargs = copy(kwargs)
        before_kwargs.pop("css_font_classes_after")
        before = html_template.render(**before_kwargs)
        before_filename = f"{filename[:-5]}-before.html"
        before_path = os.path.join(self.out, before_filename)
        with open(before_path, "w") as before_html:
            before_html.write(before)

        # This document contains fonts_after. It solely exists for
        # screenshot generation purposes
        after_kwargs = copy(kwargs)
        after_kwargs.pop("css_font_classes_before")
        after = html_template.render(**after_kwargs)
        after_filename = f"{filename[:-5]}-after.html"
        after_path = os.path.join(self.out, after_filename)
        with open(after_path, "w") as after_html:
            after_html.write(after)

        return (before_path, after_path)

    def _save_img(self, document, dst):
        # Output results as a gif
        before_page = os.path.relpath(document[0], start=self.out)
        after_page = os.path.relpath(document[1], start=self.out)
        before_url = f"{self.server_url}/{before_page}"
        after_url = f"{self.server_url}/{after_page}"
        with tempfile.TemporaryDirectory() as before_dst, tempfile.TemporaryDirectory() as after_dst:
            self.screenshot.take(before_url, before_dst)
            self.screenshot.take(after_url, after_dst)
            gen_gifs(before_dst, after_dst, dst)

    def partition(self):
        with tempfile.TemporaryDirectory() as tmp_out:
            temp_html = HtmlDiff(
                fonts_before=self.fonts_before,
                fonts_after=self.fonts_after,
                out=tmp_out,
                template_dir=self.template_dir,
            )
            # disable too_big_for_browserstack for subclasses otherwise we'll infinte loop
            temp_html.too_big_for_browserstack = False
            css_class_groups_before = partition(self.css_font_classes_before, 4)
            css_class_groups_after = partition(self.css_font_classes_after, 4)
            assert len(css_class_groups_before) == len(css_class_groups_after)
            for idx, group in enumerate(css_class_groups_before):
                temp_html.css_font_classes_before = css_class_groups_before[idx]
                temp_html.css_font_classes_after = css_class_groups_after[idx]
                for page, doc in self.documents.items():
                    temp_html.build_page(page, **doc.options)
                temp_html.save_imgs()
                src_imgs = os.path.join(temp_html.out, "img")
                dir_name = "-".join(s._style.replace(" ", "") for s in group)
                dst_imgs = os.path.join(self.out, "img", dir_name)
                shutil.rmtree(dst_imgs, ignore_errors=True)
                shutil.copytree(src_imgs, dst_imgs)


# Local server functions


def simple_server(directory="."):
    """A simple python web server which can be served from a specific
    directory

    Args:
      directory: start the server from a specified directory. Default is cwd
    """

    class Handler(SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=directory, **kwargs)

    server_address = ("", 8000)
    httpd = HTTPServer(server_address, Handler)
    httpd.serve_forever()


@contextmanager
def daemon_server(directory="."):
    """Start a simple_server as a background process.

    Args:
      directory: start the server from a specified directory. Default is '.'
    """
    p = Process(target=simple_server, args=[directory], daemon=True)
    try:
        p.start()
        yield p
    finally:
        p.terminate()


@contextmanager
def browserstack_local(access_key):
    """Start browserstack local tool as a background process"""
    # TODO This can be deprecated once
    # https://github.com/browserstack/browserstack-local-python/pull/28 is
    # merged (it may not be merged because it's a relatively inactive repo)
    local = Local(key=access_key)
    try:
        local.start()
        yield local
    finally:
        local.stop()


class ScreenShot(browserstack_screenshots.Screenshots):
    """Expansion for browserstack screenshots Lib. Adds ability to
    download files."""

    # TODO it may be better to write our own wrapper
    # since this tool isn't being developed. Code migrated from
    # https://github.com/googlefonts/diffbrowsers

    def take(self, url, dst_dir):
        """take a screenshot from a url and save it to the dst_dir"""
        self.config["url"] = url
        log.info("Taking screenshot for url: %s" % url)
        generate_resp_json = self.generate_screenshots()
        job_id = generate_resp_json["job_id"]

        log.info(
            "Browserstack is processing: "
            "http://www.browserstack.com/screenshots/%s" % job_id
        )
        screenshots_json = self.get_screenshots(job_id)
        while screenshots_json == False:  # keep refreshing until browerstack is done
            time.sleep(3)
            screenshots_json = self.get_screenshots(job_id)
        for screenshot in screenshots_json["screenshots"]:
            filename = self._build_filename_from_browserstack_json(screenshot)
            base_image = os.path.join(dst_dir, filename)
            try:
                download_file(screenshot["image_url"], base_image)
            except:
                # TODO these are too greedy
                log.warning(
                    "Skipping {} BrowserStack timed out".format(screenshot["image_url"])
                )

    def _build_filename_from_browserstack_json(self, j):
        """Build useful filename for an image from the screenshot json
        metadata"""
        filename = ""
        device = j["device"] if j["device"] else "Desktop"
        if j["state"] == "done" and j["image_url"]:
            detail = [
                device,
                j["os"],
                j["os_version"],
                j["browser"],
                j["browser_version"],
                ".png",
            ]
            filename = "_".join(item.replace(" ", "_") for item in detail if item)
        else:
            log.warning("screenshot timed out, ignoring this result")
        return filename
